@@ -27,23 +27,68 @@
 // Description:
 //    Main entry points for _omnipy Python module
 
-// $Id: omnipy.cc,v 1.1.2.25 2003/09/25 13:16:25 dgrisby Exp $
+// $Id: omnipy.cc,v 1.1.4.19 2007/09/18 20:03:34 dgrisby Exp $
 
 // $Log: omnipy.cc,v $
-// Revision 1.1.2.25  2003/09/25 13:16:25  dgrisby
-// Ensure command line arguments are strings.
+// Revision 1.1.4.19  2007/09/18 20:03:34  dgrisby
+// Refcount error if cdrUnmarshal raised a Python exception.
 //
-// Revision 1.1.2.24  2003/08/31 20:27:27  dgrisby
-// Couple of memory leaks in TypeCode unmarshalling.
+// Revision 1.1.4.18  2007/01/19 11:11:09  dgrisby
+// Avoid assertion failure if an unexpected C++ exception occurs during
+// an invocation.
 //
-// Revision 1.1.2.23  2003/07/29 14:49:28  dgrisby
-// Changes to compile with debug Python 2.3.
+// Revision 1.1.4.17  2006/09/05 11:30:22  dgrisby
+// Modify stub version check to be more specific.
 //
-// Revision 1.1.2.22  2003/07/24 15:04:34  dgrisby
-// Python 2.3's threading.Thread is a new-style class.
+// Revision 1.1.4.16  2006/09/01 13:44:06  dgrisby
+// Hard-code version check rather than using pre-processor defines.
 //
-// Revision 1.1.2.21  2003/05/28 10:13:00  dgrisby
-// Preliminary interceptor support. General clean-up.
+// Revision 1.1.4.15  2006/07/26 17:49:59  dgrisby
+// Support unchecked_narrow.
+//
+// Revision 1.1.4.14  2006/05/24 18:33:04  dgrisby
+// Unlock interpreter lock before clearing value tracker in cdrMarshal /
+// cdrUnmarshal.
+//
+// Revision 1.1.4.13  2006/05/15 10:26:11  dgrisby
+// More relaxation of requirements for old-style classes, for Python 2.5.
+//
+// Revision 1.1.4.12  2006/01/19 17:28:44  dgrisby
+// Merge from omnipy2_develop.
+//
+// Revision 1.1.4.11  2005/11/09 12:33:32  dgrisby
+// Support POA LocalObjects.
+//
+// Revision 1.1.4.10  2005/06/24 17:36:09  dgrisby
+// Support for receiving valuetypes inside Anys; relax requirement for
+// old style classes in a lot of places.
+//
+// Revision 1.1.4.9  2005/04/25 18:27:41  dgrisby
+// Maintain forwarded location when narrowing forwarded references.
+//
+// Revision 1.1.4.8  2005/04/11 13:02:38  dgrisby
+// Another merge.
+//
+// Revision 1.1.4.7  2005/03/02 13:39:17  dgrisby
+// Another merge from omnipy2_develop.
+//
+// Revision 1.1.4.6  2005/01/07 00:22:32  dgrisby
+// Big merge from omnipy2_develop.
+//
+// Revision 1.1.4.5  2004/02/16 10:14:17  dgrisby
+// Use stream based copy for local calls.
+//
+// Revision 1.1.4.4  2003/11/06 12:00:34  dgrisby
+// ValueType TypeCode support; track ORB core changes.
+//
+// Revision 1.1.4.3  2003/07/10 22:13:25  dgrisby
+// Abstract interface support.
+//
+// Revision 1.1.4.2  2003/05/20 17:10:23  dgrisby
+// Preliminary valuetype support.
+//
+// Revision 1.1.4.1  2003/03/23 21:51:57  dgrisby
+// New omnipy3_develop branch.
 //
 // Revision 1.1.2.20  2003/03/12 11:17:02  dgrisby
 // Registration of external pseudo object creation functions.
@@ -129,12 +174,21 @@ PyObject* omniPy::py_omnipymodule;	// The _omnipy extension
 PyObject* omniPy::pyCORBAmodule;	// The CORBA module
 PyObject* omniPy::pyCORBAsysExcMap;	//  The system exception map
 PyObject* omniPy::pyCORBAAnyClass;	//  Any class
+PyObject* omniPy::pyCORBATypeCodeClass;	//  TypeCode class
 PyObject* omniPy::pyCORBAContextClass;	//  Context class
+PyObject* omniPy::pyCORBAValueBase;    	//  ValueBase class
+PyObject* omniPy::pyCORBAValueBaseDesc;	//  ValueBase descriptor
 PyObject* omniPy::pyomniORBmodule;	// The omniORB module
 PyObject* omniPy::pyomniORBobjrefMap;	//  The objref class map
+PyObject* omniPy::pyomniORBskeletonMap;	//  The skeleton class map
 PyObject* omniPy::pyomniORBtypeMap;     //  The repoId to descriptor mapping
+PyObject* omniPy::pyomniORBvalueMap;    //  The repoId to value factory mapping
 PyObject* omniPy::pyomniORBwordMap;     //  Reserved word map
-PyObject* omniPy::pyPortableServerModule; // Portable server module
+PyObject* omniPy::pyomniORBpoaCache;    //  POA cache
+PyObject* omniPy::pyomniORBUnknownValueBase;
+                                        //  Base class for unknown valuetypes
+PyObject* omniPy::pyPortableServerModule;
+                                        // Portable server module
 PyObject* omniPy::pyServantClass;       // Servant class
 PyObject* omniPy::pyCreateTypeCode;	// Function to create a TypeCode object
 PyObject* omniPy::pyWorkerThreadClass;  // Worker thread class
@@ -161,8 +215,11 @@ CORBA::ORB_ptr omniPy::orb;
 // Constant strings to facilitate comparison by pointer                   //
 ////////////////////////////////////////////////////////////////////////////
 
-const char* omniPy::string_Py_omniObjRef  = "Py_omniObjRef";
-const char* omniPy::string_Py_omniServant = "Py_omniServant";
+const char* omniPy::string_Py_omniObjRef       = "Py_omniObjRef";
+const char* omniPy::string_Py_omniServant      = "Py_omniServant";
+const char* omniPy::string_Py_ServantActivator = "Py_ServantActivator";
+const char* omniPy::string_Py_ServantLocator   = "Py_ServantLocator";
+const char* omniPy::string_Py_AdapterActivator = "Py_AdapterActivator";
 
 
 ////////////////////////////////////////////////////////////////////////////
@@ -286,15 +343,16 @@ extern "C" {
     if (!PyArg_ParseTuple(args, (char*)"iis", &maj, &min, &mod))
       return 0;
 
-    if (maj != OMNIPY_MAJOR || (maj == 0 && min != OMNIPY_MINOR)) {
+    if (maj != 3 || min != 0) {
       if (omniORB::trace(1)) {
 	omniORB::logger l;
-	l << "omniORBpy: ***\n"
-	  << "omniORBpy: *** WARNING! _omnipy module version "
-	  << OMNIPY_MAJOR << "." << OMNIPY_MINOR << "\n"
-	  << "omniORBpy: *** Stubs in " << mod << " are version "
-	  << maj << "." << min << "\n"
-	  << "omniORBpy: *** You may experience strange errors "
+        l << "\n"
+	  << "omniORBpy: WARNING! _omnipy module version "
+	  << OMNIPY_MAJOR << "." << OMNIPY_MINOR
+          << " expects stubs version 3.0.\n"
+	  << "omniORBpy: Stubs in " << mod << " are version "
+	  << maj << "." << min << ".\n"
+	  << "omniORBpy: You may experience strange errors "
 	  << "until you fix the mismatch\n";
       }
     }
@@ -305,12 +363,13 @@ extern "C" {
   static PyObject*
   omnipy_coreVersion(PyObject* self, PyObject* args)
   {
-    static char* cv = (char*)OMNIORB_VERSION_STRING;
-
     if (!PyArg_ParseTuple(args, (char*)"")) return 0;
-    return Py_BuildValue((char*)"s", cv);
+    return Py_BuildValue((char*)"s", omniORB::versionString());
   }
 
+#define OMNIPY_ATTR(x) \
+  PyObject_GetAttrString(omniPy::pyomniORBmodule, (char*)x);
+
   static PyObject*
   omnipy_registerPyObjects(PyObject* self, PyObject* args)
   {
@@ -325,36 +384,36 @@ extern "C" {
 
     OMNIORB_ASSERT(PyModule_Check(omniPy::pyomniORBmodule));
 
-    omniPy::pyCORBAmodule =
-      PyObject_GetAttrString(omniPy::pyomniORBmodule, (char*)"CORBA");
+    omniPy::pyCORBAmodule = OMNIPY_ATTR("CORBA");
 
     OMNIORB_ASSERT(omniPy::pyCORBAmodule &&
 		   PyModule_Check(omniPy::pyCORBAmodule));
 
-    omniPy::pyCORBAsysExcMap =
-      PyObject_GetAttrString(omniPy::pyomniORBmodule,
-			     (char*)"sysExceptionMapping");
-
-    temp =
-      PyObject_GetAttrString(omniPy::pyCORBAmodule, (char*)"Object");
+    omniPy::pyCORBAsysExcMap = OMNIPY_ATTR("sysExceptionMapping");
 
     omniPy::pyCORBAAnyClass =
       PyObject_GetAttrString(omniPy::pyCORBAmodule, (char*)"Any");
 
+    omniPy::pyCORBATypeCodeClass =
+      PyObject_GetAttrString(omniPy::pyCORBAmodule, (char*)"TypeCode");
+
     omniPy::pyCORBAContextClass =
       PyObject_GetAttrString(omniPy::pyCORBAmodule, (char*)"Context");
 
-    omniPy::pyomniORBobjrefMap =
-      PyObject_GetAttrString(omniPy::pyomniORBmodule, (char*)"objrefMapping");
+    omniPy::pyCORBAValueBase =
+      PyObject_GetAttrString(omniPy::pyCORBAmodule, (char*)"ValueBase");
 
-    omniPy::pyomniORBtypeMap =
-      PyObject_GetAttrString(omniPy::pyomniORBmodule, (char*)"typeMapping");
+    omniPy::pyCORBAValueBaseDesc =
+      PyObject_GetAttrString(omniPy::pyCORBAmodule, (char*)"_d_ValueBase");
 
-    omniPy::pyomniORBwordMap =
-      PyObject_GetAttrString(omniPy::pyomniORBmodule, (char*)"keywordMapping");
-
-    omniPy::pyPortableServerModule =
-      PyObject_GetAttrString(omniPy::pyomniORBmodule, (char*)"PortableServer");
+    omniPy::pyomniORBobjrefMap        = OMNIPY_ATTR("objrefMapping");
+    omniPy::pyomniORBtypeMap          = OMNIPY_ATTR("typeMapping");
+    omniPy::pyomniORBwordMap          = OMNIPY_ATTR("keywordMapping");
+    omniPy::pyomniORBpoaCache         = OMNIPY_ATTR("poaCache");
+    omniPy::pyPortableServerModule    = OMNIPY_ATTR("PortableServer");
+    omniPy::pyomniORBskeletonMap      = OMNIPY_ATTR("skeletonMapping");
+    omniPy::pyomniORBvalueMap         = OMNIPY_ATTR("valueFactoryMapping");
+    omniPy::pyomniORBUnknownValueBase = OMNIPY_ATTR("UnknownValueBase");
 
     OMNIORB_ASSERT(omniPy::pyPortableServerModule);
     OMNIORB_ASSERT(PyModule_Check(omniPy::pyPortableServerModule));
@@ -362,48 +421,55 @@ extern "C" {
     omniPy::pyServantClass =
       PyObject_GetAttrString(omniPy::pyPortableServerModule, (char*)"Servant");
 
-    temp =
-      PyObject_GetAttrString(omniPy::pyomniORBmodule, (char*)"tcInternal");
+    temp = OMNIPY_ATTR("tcInternal");
 
     omniPy::pyCreateTypeCode = PyObject_GetAttrString(temp,
 						      (char*)"createTypeCode");
 
-    omniPy::pyWorkerThreadClass =
-      PyObject_GetAttrString(omniPy::pyomniORBmodule, (char*)"WorkerThread");
+    omniPy::pyWorkerThreadClass = OMNIPY_ATTR("WorkerThread");
 
     omniPy::pyWorkerThreadDel =
       PyObject_GetAttrString(omniPy::pyWorkerThreadClass, (char*)"delete");
 
-    omniPy::pyEmptyTuple = PyTuple_New(0);
+    omniPy::pyEmptyTuple = OMNIPY_ATTR("_emptyTuple");
 
     OMNIORB_ASSERT(omniPy::pyCORBAsysExcMap);
     OMNIORB_ASSERT(PyDict_Check(omniPy::pyCORBAsysExcMap));
     OMNIORB_ASSERT(omniPy::pyCORBAAnyClass);
-    OMNIORB_ASSERT(PyClass_Check(omniPy::pyCORBAAnyClass));
+    OMNIORB_ASSERT(omniPy::pyCORBATypeCodeClass);
     OMNIORB_ASSERT(omniPy::pyCORBAContextClass);
-    OMNIORB_ASSERT(PyClass_Check(omniPy::pyCORBAContextClass));
+    OMNIORB_ASSERT(omniPy::pyCORBAValueBaseDesc);
+    OMNIORB_ASSERT(PyTuple_Check(omniPy::pyCORBAValueBaseDesc));
+    OMNIORB_ASSERT(omniPy::pyCORBAValueBase);
     OMNIORB_ASSERT(omniPy::pyomniORBobjrefMap);
     OMNIORB_ASSERT(PyDict_Check(omniPy::pyomniORBobjrefMap));
+    OMNIORB_ASSERT(omniPy::pyomniORBskeletonMap);
+    OMNIORB_ASSERT(PyDict_Check(omniPy::pyomniORBskeletonMap));
     OMNIORB_ASSERT(omniPy::pyomniORBtypeMap);
     OMNIORB_ASSERT(PyDict_Check(omniPy::pyomniORBtypeMap));
+    OMNIORB_ASSERT(omniPy::pyomniORBvalueMap);
+    OMNIORB_ASSERT(PyDict_Check(omniPy::pyomniORBvalueMap));
     OMNIORB_ASSERT(omniPy::pyomniORBwordMap);
     OMNIORB_ASSERT(PyDict_Check(omniPy::pyomniORBwordMap));
+    OMNIORB_ASSERT(omniPy::pyomniORBUnknownValueBase);
+    OMNIORB_ASSERT(omniPy::pyomniORBpoaCache);
+    OMNIORB_ASSERT(PyDict_Check(omniPy::pyomniORBpoaCache));
     OMNIORB_ASSERT(omniPy::pyServantClass);
-    OMNIORB_ASSERT(PyClass_Check(omniPy::pyServantClass));
     OMNIORB_ASSERT(omniPy::pyCreateTypeCode);
     OMNIORB_ASSERT(PyFunction_Check(omniPy::pyCreateTypeCode));
     OMNIORB_ASSERT(omniPy::pyWorkerThreadClass);
     OMNIORB_ASSERT(omniPy::pyWorkerThreadDel);
     OMNIORB_ASSERT(PyMethod_Check(omniPy::pyWorkerThreadDel));
     OMNIORB_ASSERT(omniPy::pyEmptyTuple);
+    OMNIORB_ASSERT(PyTuple_Check(omniPy::pyEmptyTuple));
 
-    omniPy::pyORB_TWIN        = PyString_FromString((char*)"__omni_orb");
-    omniPy::pyOBJREF_TWIN     = PyString_FromString((char*)"__omni_obj");
-    omniPy::pySERVANT_TWIN    = PyString_FromString((char*)"__omni_svt");
-    omniPy::pyPOA_TWIN        = PyString_FromString((char*)"__omni_poa");
-    omniPy::pyPOAMANAGER_TWIN = PyString_FromString((char*)"__omni_mgr");
-    omniPy::pyPOACURRENT_TWIN = PyString_FromString((char*)"__omni_pct");
-    omniPy::pyNP_RepositoryId = PyString_FromString((char*)"_NP_RepositoryId");
+    omniPy::pyORB_TWIN        = OMNIPY_ATTR("_ORB_TWIN");
+    omniPy::pyOBJREF_TWIN     = OMNIPY_ATTR("_OBJREF_TWIN");
+    omniPy::pySERVANT_TWIN    = OMNIPY_ATTR("_SERVANT_TWIN");
+    omniPy::pyPOA_TWIN        = OMNIPY_ATTR("_POA_TWIN");
+    omniPy::pyPOAMANAGER_TWIN = OMNIPY_ATTR("_POAMANAGER_TWIN");
+    omniPy::pyPOACURRENT_TWIN = OMNIPY_ATTR("_POACURRENT_TWIN");
+    omniPy::pyNP_RepositoryId = OMNIPY_ATTR("_NP_RepositoryId");
 
     OMNIORB_ASSERT(omniPy::pyORB_TWIN);
     OMNIORB_ASSERT(omniPy::pyOBJREF_TWIN);
@@ -412,6 +478,13 @@ extern "C" {
     OMNIORB_ASSERT(omniPy::pyPOAMANAGER_TWIN);
     OMNIORB_ASSERT(omniPy::pyPOACURRENT_TWIN);
 
+    OMNIORB_ASSERT(PyString_Check(omniPy::pyORB_TWIN));
+    OMNIORB_ASSERT(PyString_Check(omniPy::pyOBJREF_TWIN));
+    OMNIORB_ASSERT(PyString_Check(omniPy::pySERVANT_TWIN));
+    OMNIORB_ASSERT(PyString_Check(omniPy::pyPOA_TWIN));
+    OMNIORB_ASSERT(PyString_Check(omniPy::pyPOAMANAGER_TWIN));
+    OMNIORB_ASSERT(PyString_Check(omniPy::pyPOACURRENT_TWIN));
+
     Py_INCREF(Py_None);
     return Py_None;
   }
@@ -523,39 +596,109 @@ extern "C" {
   omnipy_cdrMarshal(PyObject* self, PyObject* args)
   {
     PyObject *desc, *data;
+    int endian = -1;
 
-    if (!PyArg_ParseTuple(args, (char*)"OO", &desc, &data)) return 0;
+    if (!PyArg_ParseTuple(args, (char*)"OO|i", &desc, &data, &endian))
+      return 0;
+
+    if (endian > 1 || endian < -1) {
+      PyErr_SetString(PyExc_ValueError,
+		      "argument 3: endian must be 0 or 1");
+      return 0;
+    }
 
     try {
-      cdrEncapsulationStream stream;
-
       omniPy::validateType(desc, data, CORBA::COMPLETED_NO);
-      omniPy::marshalPyObject(stream, desc, data);
 
-      return PyString_FromStringAndSize((char*)stream.bufPtr(),
-					stream.bufSize());
+      if (endian == -1) {
+	// Marshal into an encapsulation
+	cdrEncapsulationStream stream;
+        omniPy::ValueTrackerClearer vtc(stream);
+
+	omniPy::marshalPyObject(stream, desc, data);
+
+	return PyString_FromStringAndSize((char*)stream.bufPtr(),
+					  stream.bufSize());
+      }
+      else {
+	// Marshal into a raw buffer
+	cdrMemoryStream stream;
+        omniPy::ValueTrackerClearer vtc(stream);
+
+	if (endian != omni::myByteOrder)
+	  stream.setByteSwapFlag(endian);
+
+	omniPy::marshalPyObject(stream, desc, data);
+
+	return PyString_FromStringAndSize((char*)stream.bufPtr(),
+					  stream.bufSize());
+      }
     }
     OMNIPY_CATCH_AND_HANDLE_SYSTEM_EXCEPTIONS
   }
 
+  static inline PyObject* do_cdrUnmarshal(cdrStream& stream, PyObject* desc)
+  {
+    PyObject* r = omniPy::unmarshalPyObject(stream, desc);
+    if (r && stream.checkInputOverrun(1, 1)) {
+      // More data in stream -- must have used the wrong TypeCode
+      Py_DECREF(r);
+      OMNIORB_THROW(MARSHAL, MARSHAL_MessageTooLong, CORBA::COMPLETED_NO);
+    }
+    return r;
+  }
+
   static PyObject*
   omnipy_cdrUnmarshal(PyObject* self, PyObject* args)
   {
     PyObject* desc;
     char*     encap;
     size_t    size;
+    int       endian = -1;
     
-    if (!PyArg_ParseTuple(args, (char*)"Os#", &desc, &encap, &size)) return 0;
+    if (!PyArg_ParseTuple(args, (char*)"Os#|i",
+			  &desc, &encap, &size, &endian))
+      return 0;
+
+    if (endian > 1 || endian < -1) {
+      PyErr_SetString(PyExc_ValueError,
+		      "argument 3: endian must be 0 or 1");
+      return 0;
+    }
 
     try {
-      cdrEncapsulationStream stream((CORBA::Octet*)encap, size);
-      PyObject* r = omniPy::unmarshalPyObject(stream, desc);
-      if (stream.checkInputOverrun(1, 1)) {
-	// More data in stream -- must have used the wrong TypeCode
-	Py_DECREF(r);
-	OMNIORB_THROW(MARSHAL, MARSHAL_MessageTooLong, CORBA::COMPLETED_NO);
+      if (endian == -1) {
+	// Encapsulation
+	cdrEncapsulationStream stream((CORBA::Octet*)encap, size);
+        omniPy::ValueTrackerClearer vtc(stream);
+	return do_cdrUnmarshal(stream, desc);
+      }
+      else {
+	// Simple buffer. Is it aligned ok?
+	if ((omni::ptr_arith_t)encap ==
+	    omni::align_to((omni::ptr_arith_t)encap, omni::ALIGN_8)) {
+
+	  cdrMemoryStream stream((CORBA::Octet*)encap, size);
+          omniPy::ValueTrackerClearer vtc(stream);
+
+	  if (endian != omni::myByteOrder)
+	    stream.setByteSwapFlag(endian);
+
+	  return do_cdrUnmarshal(stream, desc);
+	}
+	else {
+	  // Unfortunately, this is a common case, due to the way
+	  // Python string objects are laid out.
+	  cdrMemoryStream stream;
+          omniPy::ValueTrackerClearer vtc(stream);
+
+	  if (endian != omni::myByteOrder)
+	    stream.setByteSwapFlag(endian);
+
+	  stream.put_octet_array((CORBA::Octet*)encap, size);
+	  return do_cdrUnmarshal(stream, desc);
+	}
       }
-      return r;
     }
     OMNIPY_CATCH_AND_HANDLE_SYSTEM_EXCEPTIONS
   }
@@ -569,7 +712,7 @@ extern "C" {
   omnipy_invoke(PyObject* self, PyObject* args)
   {
     // Arg format
-    //  (objref, op_name, (in_desc, out_desc, exc_desc), args)
+    //  (objref, op_name, (in_desc,out_desc,exc_desc [, ctxt [,values]]), args)
     //
     //  exc_desc is a dictionary containing a mapping from repoIds to
     //  tuples of the form (exception class, marshal desc., param count)
@@ -578,11 +721,6 @@ extern "C" {
     char*  op;
     size_t op_len;
 
-    /*
-    if (!PyArg_ParseTuple(args, "Os#(OOO)O", &pyobjref, &op, &op_len,
-			  &in_d, &out_d, &exc_d, &op_args))
-      return 0;
-    */
     PyObject* o;
     PyObject* desc;
 
@@ -598,11 +736,24 @@ extern "C" {
     out_d    = PyTuple_GET_ITEM(desc,1);
     exc_d    = PyTuple_GET_ITEM(desc,2);
 
-    if (PyTuple_GET_SIZE(desc) == 4)
+    int desclen = PyTuple_GET_SIZE(desc);
+
+    if (desclen >= 4) {
       ctxt_d = PyTuple_GET_ITEM(desc,3);
+      if (ctxt_d == Py_None)
+	ctxt_d = 0;
+    }
     else
       ctxt_d = 0;
 
+    CORBA::Boolean contains_values = 0;
+
+    if (desclen == 5) {
+      PyObject* v = PyTuple_GET_ITEM(desc,4);
+      if (v != Py_None)
+	contains_values = 1;
+    }
+
     op_args  = PyTuple_GET_ITEM(args,3);
 
     int arg_len = PyTuple_GET_SIZE(in_d) + (ctxt_d ? 1:0);
@@ -627,6 +778,10 @@ extern "C" {
     omniPy::Py_omniCallDescriptor call_desc(op, op_len + 1, is_oneway,
 					    in_d, out_d, exc_d, ctxt_d,
 					    op_args, 0);
+
+    if (contains_values)
+      call_desc.containsValues(1);
+
     try {
       call_desc.releaseInterpreterLock();
       oobjref->_invoke(call_desc);
@@ -656,6 +811,14 @@ OMNIORB_FOR_EACH_SYS_EXCEPTION(DO_CALL_DESC_SYSTEM_EXCEPTON)
       call_desc.reacquireInterpreterLock();
       ex.setPyExceptionState();
     }
+    catch (...) {
+      // This should not happen, but in case it does we reacquire the
+      // interpreter lock to avoid an assertion failure from the call
+      // descriptor's destructor.
+      omniORB::logs(1, "Unexpected C++ exception during Python invocation.");
+      call_desc.ensureInterpreterLock();
+      throw;
+    }
     return 0;
   }
 
@@ -667,8 +830,6 @@ OMNIORB_FOR_EACH_SYS_EXCEPTION(DO_CALL_DESC_SYSTEM_EXCEPTON)
     if (!PyArg_ParseTuple(args, (char*)"O", &pyobjref))
       return 0;
 
-    OMNIORB_ASSERT(PyInstance_Check(pyobjref));
-
     CORBA::Object_ptr cxxobjref =
       (CORBA::Object_ptr)omniPy::getTwin(pyobjref, OBJREF_TWIN);
 
@@ -692,9 +853,6 @@ OMNIORB_FOR_EACH_SYS_EXCEPTION(DO_CALL_DESC_SYSTEM_EXCEPTON)
     if (!PyArg_ParseTuple(args, (char*)"Os", &pyobjref, &repoId))
       return 0;
 
-    RAISE_PY_BAD_PARAM_IF(!PyInstance_Check(pyobjref),
-			  BAD_PARAM_WrongPythonType);
-
     CORBA::Object_ptr cxxobjref =
       (CORBA::Object_ptr)omniPy::getTwin(pyobjref, OBJREF_TWIN);
 
@@ -716,9 +874,6 @@ OMNIORB_FOR_EACH_SYS_EXCEPTION(DO_CALL_DESC_SYSTEM_EXCEPTON)
     if (!PyArg_ParseTuple(args, (char*)"O", &pyobjref))
       return 0;
 
-    RAISE_PY_BAD_PARAM_IF(!PyInstance_Check(pyobjref),
-			  BAD_PARAM_WrongPythonType);
-
     CORBA::Object_ptr cxxobjref =
       (CORBA::Object_ptr)omniPy::getTwin(pyobjref, OBJREF_TWIN);
 
@@ -741,10 +896,6 @@ OMNIORB_FOR_EACH_SYS_EXCEPTION(DO_CALL_DESC_SYSTEM_EXCEPTON)
     if (!PyArg_ParseTuple(args, (char*)"OO", &pyobjref1, &pyobjref2))
       return 0;
 
-    RAISE_PY_BAD_PARAM_IF(!PyInstance_Check(pyobjref1) ||
-			  !PyInstance_Check(pyobjref2),
-			  BAD_PARAM_WrongPythonType);
-
     CORBA::Object_ptr cxxobjref1, cxxobjref2;
     cxxobjref1 = (CORBA::Object_ptr)omniPy::getTwin(pyobjref1, OBJREF_TWIN);
     cxxobjref2 = (CORBA::Object_ptr)omniPy::getTwin(pyobjref2, OBJREF_TWIN);
@@ -769,9 +920,6 @@ OMNIORB_FOR_EACH_SYS_EXCEPTION(DO_CALL_DESC_SYSTEM_EXCEPTON)
     if (!PyArg_ParseTuple(args, (char*)"Oi", &pyobjref, &max))
       return 0;
 
-    RAISE_PY_BAD_PARAM_IF(!PyInstance_Check(pyobjref),
-			  BAD_PARAM_WrongPythonType);
-
     CORBA::Object_ptr cxxobjref =
       (CORBA::Object_ptr)omniPy::getTwin(pyobjref, OBJREF_TWIN);
 
@@ -787,13 +935,11 @@ OMNIORB_FOR_EACH_SYS_EXCEPTION(DO_CALL_DESC_SYSTEM_EXCEPTON)
   {
     PyObject* pysource;
     char*     repoId;
+    int       checked;
 
-    if (!PyArg_ParseTuple(args, (char*)"Os", &pysource, &repoId))
+    if (!PyArg_ParseTuple(args, (char*)"Osi", &pysource, &repoId, &checked))
       return 0;
 
-    RAISE_PY_BAD_PARAM_IF(!PyInstance_Check(pysource),
-			  BAD_PARAM_WrongPythonType);
-
     CORBA::Object_ptr cxxsource =
       (CORBA::Object_ptr)omniPy::getTwin(pysource, OBJREF_TWIN);
 
@@ -804,14 +950,22 @@ OMNIORB_FOR_EACH_SYS_EXCEPTION(DO_CALL_DESC_SYSTEM_EXCEPTON)
 
     try {
       omniPy::InterpreterUnlocker ul;
-      isa = cxxsource->_is_a(repoId);
+
+      if (checked || cxxsource->_NP_is_pseudo())
+	isa = cxxsource->_is_a(repoId);
+      else
+	isa = 1;
 
       if (isa) {
 	if (!cxxsource->_NP_is_pseudo()) {
 	  omniObjRef* oosource = cxxsource->_PR_getobj();
-	  omniObjRef* oodest   = omniPy::createObjRef(repoId,
-						      oosource->_getIOR(),
-						      0, 0, 1);
+	  omniObjRef* oodest;
+	  {
+	    omni_tracedmutex_lock sync(*omni::internalLock);
+	    oodest = omniPy::createObjRef(repoId, oosource->_getIOR(), 1,
+					  oosource->_identity(), 1,
+					  oosource->_isForwardLocation());
+	  }
 	  cxxdest = (CORBA::Object_ptr)
 	                   (oodest->_ptrToObjRef(CORBA::Object::_PD_repoId));
 	}
@@ -830,6 +984,33 @@ OMNIORB_FOR_EACH_SYS_EXCEPTION(DO_CALL_DESC_SYSTEM_EXCEPTON)
     }
   }
 
+  static PyObject*
+  omnipy_ensureInit(PyObject* self, PyObject* args)
+  {
+    PyObject* m = PyImport_ImportModule((char*)"_omnipy");
+    PyObject* o = PyObject_GetAttrString(m, (char*)"orb_func");
+    PyObject* f = 0;
+    
+    if (o && PyModule_Check(o))
+      f = PyObject_GetAttrString(o, (char*)"destroy");
+
+    if (!o || !PyModule_Check(o) || !f || f == Py_None) {
+      omniORB::logs(5, "Reinitialise omniORBpy sub-modules.");
+      PyObject* d = PyModule_GetDict(m);
+      omniPy::initORBFunc(d);
+      omniPy::initPOAFunc(d);
+      omniPy::initPOAManagerFunc(d);
+      omniPy::initPOACurrentFunc(d);
+      omniPy::initInterceptorFunc(d);
+      omniPy::initomniFunc(d);
+    }
+    Py_XDECREF(o);
+    Py_XDECREF(f);
+
+    Py_INCREF(Py_None);
+    return Py_None;
+  }
+
 
   ////////////////////////////////////////////////////////////////////////////
   // Python method table                                                    //
@@ -844,6 +1025,7 @@ OMNIORB_FOR_EACH_SYS_EXCEPTION(DO_CALL_DESC_SYSTEM_EXCEPTON)
     {(char*)"cdrMarshal",        omnipy_cdrMarshal,              METH_VARARGS},
     {(char*)"cdrUnmarshal",      omnipy_cdrUnmarshal,            METH_VARARGS},
     {(char*)"need_ORB_init",     omnipy_need_ORB_init,           METH_VARARGS},
+    {(char*)"ensureInit",        omnipy_ensureInit,              METH_VARARGS},
 
     // Wrappers for functions in CORBA::
     {(char*)"ORB_init",          omnipy_ORB_init,                METH_VARARGS},
