@@ -31,8 +31,47 @@
 #define _omnipy_h_
 
 // $Log: omnipy.h,v $
-// Revision 1.2.4.21  2003/05/28 10:13:01  dgrisby
-// Preliminary interceptor support. General clean-up.
+// Revision 1.3.2.13  2007/01/19 11:11:09  dgrisby
+// Avoid assertion failure if an unexpected C++ exception occurs during
+// an invocation.
+//
+// Revision 1.3.2.12  2006/07/26 17:50:43  dgrisby
+// Reuse existing omniIOR object when converting C++ object reference to Python.
+//
+// Revision 1.3.2.11  2006/07/19 09:40:39  dgrisby
+// Track ORB core changes.
+//
+// Revision 1.3.2.10  2006/05/24 18:33:04  dgrisby
+// Unlock interpreter lock before clearing value tracker in cdrMarshal /
+// cdrUnmarshal.
+//
+// Revision 1.3.2.9  2006/05/15 10:26:11  dgrisby
+// More relaxation of requirements for old-style classes, for Python 2.5.
+//
+// Revision 1.3.2.8  2006/01/19 17:28:44  dgrisby
+// Merge from omnipy2_develop.
+//
+// Revision 1.3.2.7  2005/11/09 12:33:32  dgrisby
+// Support POA LocalObjects.
+//
+// Revision 1.3.2.6  2005/06/24 17:36:08  dgrisby
+// Support for receiving valuetypes inside Anys; relax requirement for
+// old style classes in a lot of places.
+//
+// Revision 1.3.2.5  2005/04/25 18:27:41  dgrisby
+// Maintain forwarded location when narrowing forwarded references.
+//
+// Revision 1.3.2.4  2005/01/07 00:22:32  dgrisby
+// Big merge from omnipy2_develop.
+//
+// Revision 1.3.2.3  2003/07/10 22:13:25  dgrisby
+// Abstract interface support.
+//
+// Revision 1.3.2.2  2003/05/20 17:10:23  dgrisby
+// Preliminary valuetype support.
+//
+// Revision 1.3.2.1  2003/03/23 21:51:57  dgrisby
+// New omnipy3_develop branch.
 //
 // Revision 1.2.4.20  2003/03/12 11:17:02  dgrisby
 // Registration of external pseudo object creation functions.
@@ -145,21 +184,30 @@ public:
   // Global pointers to Python objects                                      //
   ////////////////////////////////////////////////////////////////////////////
 
-  static PyObject* py_omnipymodule;    // _omnipy module
-  static PyObject* pyCORBAmodule;      // CORBA module
-  static PyObject* pyCORBAsysExcMap;   //  The system exception map
-  static PyObject* pyCORBAAnyClass;    //  Any class
-  static PyObject* pyCORBAContextClass;//  Context class
-  static PyObject* pyomniORBmodule;    // The omniORB module
-  static PyObject* pyomniORBobjrefMap; //  The objref class map
-  static PyObject* pyomniORBtypeMap;   //  Type map
-  static PyObject* pyomniORBwordMap;   //  Reserved word map
-  static PyObject* pyPortableServerModule; // Portable server module
-  static PyObject* pyServantClass;     // Servant class
-  static PyObject* pyCreateTypeCode;   // Function to create a TypeCode object
-  static PyObject* pyWorkerThreadClass;// Worker thread class
-  static PyObject* pyWorkerThreadDel;  // Method to delete worker thread
-  static PyObject* pyEmptyTuple;       // Zero element tuple
+  static PyObject* py_omnipymodule;    	// _omnipy module
+  static PyObject* pyCORBAmodule;      	// CORBA module
+  static PyObject* pyCORBAsysExcMap;   	//  The system exception map
+  static PyObject* pyCORBAAnyClass;    	//  Any class
+  static PyObject* pyCORBATypeCodeClass;//  TypeCode class
+  static PyObject* pyCORBAContextClass;	//  Context class
+  static PyObject* pyCORBAValueBase;    //  ValueBase class
+  static PyObject* pyCORBAValueBaseDesc;//  Descriptor for ValueBase
+  static PyObject* pyomniORBmodule;    	// The omniORB module
+  static PyObject* pyomniORBobjrefMap; 	//  The objref class map
+  static PyObject* pyomniORBskeletonMap;//  The skeleton class map
+  static PyObject* pyomniORBtypeMap;   	//  Type map
+  static PyObject* pyomniORBvalueMap;  	//  Value factory map
+  static PyObject* pyomniORBwordMap;   	//  Reserved word map
+  static PyObject* pyomniORBpoaCache;   //  POA cache
+  static PyObject* pyomniORBUnknownValueBase;
+                                        //  Base class for unknown valuetypes
+  static PyObject* pyPortableServerModule;
+                                        // Portable server module
+  static PyObject* pyServantClass;     	// Servant class
+  static PyObject* pyCreateTypeCode;   	// Function to create a TypeCode object
+  static PyObject* pyWorkerThreadClass;	// Worker thread class
+  static PyObject* pyWorkerThreadDel;  	// Method to delete worker thread
+  static PyObject* pyEmptyTuple;       	// Zero element tuple
 
   ////////////////////////////////////////////////////////////////////////////
   // Twin names and other 'static' strings                                  //
@@ -179,6 +227,9 @@ public:
 
   static const char* string_Py_omniObjRef;
   static const char* string_Py_omniServant;
+  static const char* string_Py_ServantActivator;
+  static const char* string_Py_ServantLocator;
+  static const char* string_Py_AdapterActivator;
 
 
   ////////////////////////////////////////////////////////////////////////////
@@ -189,7 +240,7 @@ public:
 
 
   ////////////////////////////////////////////////////////////////////////////
-  // C++ API objrect                                                        //
+  // C++ API object                                                         //
   ////////////////////////////////////////////////////////////////////////////
 
   static omniORBpyAPI cxxAPI;
@@ -206,8 +257,15 @@ public:
   setTwin(PyObject* obj, void* twin, PyObject* name)
   {
     PyObject* ot = newTwin(twin);
+    PyObject_SetAttr(obj, name, ot);
+    Py_DECREF(ot);
+  }
 
-    PyDict_SetItem(((PyInstanceObject*)obj)->in_dict, name, ot);
+  static
+  inline void
+  setExistingTwin(PyObject* obj, PyObject* ot, PyObject* name)
+  {
+    PyObject_SetAttr(obj, name, ot);
     Py_DECREF(ot);
   }
 
@@ -215,18 +273,24 @@ public:
   inline void*
   getTwin(PyObject* obj, PyObject* name)
   {
-    PyObject* ot = PyDict_GetItem(((PyInstanceObject*)obj)->in_dict, name);
-    if (ot)
-      return ((omnipyTwin*)ot)->ob_twin;
-    else
-      return 0;
+    void* twin;
+    PyObject* ot = PyObject_GetAttr(obj, name);
+    if (ot) {
+      twin = ((omnipyTwin*)ot)->ob_twin;
+      Py_DECREF(ot);
+    }
+    else {
+      PyErr_Clear();
+      twin = 0;
+    }
+    return twin;
   }
 
   static
   inline void
   remTwin(PyObject* obj, PyObject* name)
   {
-    PyDict_DelItem(((PyInstanceObject*)obj)->in_dict, name);
+    PyObject_DelAttr(obj, name);
   }
 
   ////////////////////////////////////////////////////////////////////////////
@@ -257,7 +321,8 @@ public:
 
   // Throw a C++ system exception equivalent to the given Python exception
   static
-  void produceSystemException(PyObject* eobj, PyObject* erepoId);
+  void produceSystemException(PyObject* eobj, PyObject* erepoId,
+			      PyObject* etype, PyObject* etraceback);
 
   // Handle the current Python exception. An exception must have
   // occurred. Deals with system exceptions and
@@ -270,11 +335,33 @@ public:
   static
   void handleLocationForward(PyObject* evalue);
   
-
   // Ensure there is an omni_thread associated with the calling thread.
   static
   omni_thread* ensureOmniThread();
-  
+
+  // IsInstance function for all Python versions.
+  static inline
+  CORBA::Boolean isInstance(PyObject* o, PyObject* c)
+  {
+#if PY_VERSION_HEX >= 0x02010000
+    return PyObject_IsInstance(o,c);
+#else
+    if (!PyInstance_Check(a_o)) return 0;
+    PyObject* acls = (PyObject*)((PyInstanceObject*)o)->in_class;
+    return PyClass_IsSubclass(acls, c);
+#endif
+  }
+
+  // IsSubclass function for all Python versions
+  static inline
+  CORBA::Boolean isSubclass(PyObject* o, PyObject* c)
+  {
+#if PY_VERSION_HEX >= 0x02010000
+    return PyObject_IsSubclass(o,c);
+#else
+    return PyClass_IsSubclass(o, c);
+#endif
+  }
 
   ////////////////////////////////////////////////////////////////////////////
   // Fixed point                                                            //
@@ -327,19 +414,22 @@ public:
 			   omniIOR*           ior,
 			   CORBA::Boolean     locked,
 			   omniIdentity*      id = 0,
-			   CORBA::Boolean     type_verified = 0);
+			   CORBA::Boolean     type_verified = 0,
+			   CORBA::Boolean     is_forwarded = 0);
 
   static
-  omniObjRef* createLocalObjRef(const char*        mostDerivedRepoId,
-				const char*        targetRepoId,
-				omniObjTableEntry* entry,
-				CORBA::Boolean     type_verified = 0);
+  omniObjRef* createLocalObjRef(const char*         mostDerivedRepoId,
+				const char*         targetRepoId,
+				omniObjTableEntry*  entry,
+				omniObjRef*         orig_ref,
+				CORBA::Boolean      type_verified = 0);
 
   static
   omniObjRef* createLocalObjRef(const char* 	    mostDerivedRepoId,
 				const char* 	    targetRepoId,
 				const _CORBA_Octet* key,
 				int                 keysize,
+				omniObjRef*         orig_ref,
 				CORBA::Boolean      type_verified = 0);
 
   // When a POA creates a reference to a Python servant, it does not
@@ -394,24 +484,27 @@ public:
   // completion status.
   //
   typedef void (*ValidateTypeFn)(PyObject*, PyObject*,
-				 CORBA::CompletionStatus);
+				 CORBA::CompletionStatus,
+				 PyObject*);
 
   static const ValidateTypeFn validateTypeFns[];
 
   static void validateTypeIndirect(PyObject* d_o, PyObject* a_o,
-				   CORBA::CompletionStatus compstatus);
+				   CORBA::CompletionStatus compstatus,
+				   PyObject* track);
 
   static inline
   void validateType(PyObject* d_o, PyObject* a_o,
-		    CORBA::CompletionStatus compstatus)
+		    CORBA::CompletionStatus compstatus,
+		    PyObject* track = 0)
   {
     CORBA::ULong tk = descriptorToTK(d_o);
 
     if (tk <= 33) { // tk_local_interface
-      validateTypeFns[tk](d_o, a_o, compstatus);
+      validateTypeFns[tk](d_o, a_o, compstatus, track);
     }
     else if (tk == 0xffffffff) { // Indirection
-      validateTypeIndirect(d_o, a_o, compstatus);
+      validateTypeIndirect(d_o, a_o, compstatus, track);
     }
     else OMNIORB_THROW(BAD_TYPECODE, BAD_TYPECODE_UnknownKind, compstatus);
   }
@@ -521,6 +614,56 @@ public:
     return pystring;
   }
 
+
+  ////////////////////////////////////////////////////////////////////////////
+  // Valuetype / abstract interface marshalling functions                   //
+  ////////////////////////////////////////////////////////////////////////////
+
+  static void
+  validateTypeValue(PyObject* d_o, PyObject* a_o,
+		    CORBA::CompletionStatus compstatus,
+		    PyObject* track);
+
+  static void
+  validateTypeValueBox(PyObject* d_o, PyObject* a_o,
+		       CORBA::CompletionStatus compstatus,
+		       PyObject* track);
+
+  static void
+  validateTypeAbstractInterface(PyObject* d_o, PyObject* a_o,
+				CORBA::CompletionStatus compstatus,
+				PyObject* track);
+
+  static void
+  marshalPyObjectValue(cdrStream& stream, PyObject* d_o, PyObject* a_o);
+
+  static void
+  marshalPyObjectValueBox(cdrStream& stream, PyObject* d_o, PyObject* a_o);
+
+  static void
+  marshalPyObjectAbstractInterface(cdrStream& stream,
+				   PyObject* d_o, PyObject* a_o);
+
+  static PyObject*
+  unmarshalPyObjectValue(cdrStream& stream, PyObject* d_o);
+  // Shared by Value and ValueBox
+
+  static PyObject*
+  unmarshalPyObjectAbstractInterface(cdrStream& stream, PyObject* d_o);
+
+  static PyObject*
+  copyArgumentValue(PyObject* d_o, PyObject* a_o,
+		    CORBA::CompletionStatus compstatus);
+
+  static PyObject*
+  copyArgumentValueBox(PyObject* d_o, PyObject* a_o,
+		       CORBA::CompletionStatus compstatus);
+
+  static PyObject*
+  copyArgumentAbstractInterface(PyObject* d_o, PyObject* a_o,
+				CORBA::CompletionStatus compstatus);
+
+
   ////////////////////////////////////////////////////////////////////////////
   // TypeCode and Any support functions                                     //
   ////////////////////////////////////////////////////////////////////////////
@@ -619,8 +762,16 @@ public:
       tstate_ = 0;
     }
 
-    // Extract and take ownership of stored args/results
-    inline PyObject* args()   { PyObject* r = args_;   args_ = 0;   return r; }
+    inline void ensureInterpreterLock() {
+      if (tstate_) {
+        PyEval_RestoreThread(tstate_);
+        tstate_ = 0;
+      }
+    }
+
+    inline PyObject* args() { return args_; }
+
+    // Extract and take ownership of stored results
     inline PyObject* result() { PyObject* r = result_; result_ = 0; return r; }
 
     //
@@ -728,6 +879,95 @@ public:
   static Py_omniServant* getServantForPyObject(PyObject* pyservant);
 
 
+  ////////////////////////////////////////////////////////////////////////////
+  // ServantManager / AdapterActivator implementations                      //
+  ////////////////////////////////////////////////////////////////////////////
+
+  class Py_ServantActivator
+  {
+  public:
+    Py_ServantActivator(PyObject* pysa) : pysa_(pysa) { Py_INCREF(pysa_); }
+    ~Py_ServantActivator() { Py_DECREF(pysa_); }
+
+    PortableServer::Servant incarnate(const PortableServer::ObjectId& oid,
+				      PortableServer::POA_ptr         poa);
+
+    void etherealize(const PortableServer::ObjectId& oid,
+		     PortableServer::POA_ptr         poa,
+		     PortableServer::Servant         serv,
+		     CORBA::Boolean                  cleanup_in_progress,
+		     CORBA::Boolean                  remaining_activations);
+
+    inline PyObject* pyobj() { return pysa_; }
+
+  private:
+    PyObject* pysa_;
+
+    // Not implemented
+    Py_ServantActivator(const Py_ServantActivator&);
+    Py_ServantActivator& operator=(const Py_ServantActivator&);
+  };
+
+  class Py_ServantLocator
+  {
+  public:
+    Py_ServantLocator(PyObject* pysl) : pysl_(pysl) { Py_INCREF(pysl_); }
+    ~Py_ServantLocator() { Py_DECREF(pysl_); }
+
+    PortableServer::Servant preinvoke(const PortableServer::ObjectId& oid,
+				      PortableServer::POA_ptr poa,
+				      const char*             operation,
+				      void*&                  cookie);
+
+    void postinvoke(const PortableServer::ObjectId& oid,
+		    PortableServer::POA_ptr         poa,
+		    const char*                     operation,
+		    void*                           cookie,
+		    PortableServer::Servant         serv);
+
+    inline PyObject* pyobj() { return pysl_; }
+
+  private:
+    PyObject* pysl_;
+
+    // Not implemented
+    Py_ServantLocator(const Py_ServantLocator&);
+    Py_ServantLocator& operator=(const Py_ServantLocator&);
+  };
+
+  class Py_AdapterActivator
+  {
+  public:
+    Py_AdapterActivator(PyObject* pyaa) : pyaa_(pyaa) { Py_INCREF(pyaa_); }
+    ~Py_AdapterActivator() { Py_DECREF(pyaa_); }
+
+    CORBA::Boolean unknown_adapter(PortableServer::POA_ptr parent,
+				   const char*             name);
+
+    inline PyObject* pyobj() { return pyaa_; }
+
+  private:
+    PyObject* pyaa_;
+
+    // Not implemented
+    Py_AdapterActivator(const Py_AdapterActivator&);
+    Py_AdapterActivator& operator=(const Py_AdapterActivator&);
+  };
+
+  // Function to create a C++ local object for a Python object. If the
+  // Python object is not an instance of a mapped local object,
+  // returns 0.
+  //
+  // Caller must hold the Python interpreter lock.
+  static CORBA::LocalObject_ptr getLocalObjectForPyObject(PyObject* pyobj);
+
+  // Convert a LocalObject to the underlying Python object. If the
+  // object is not a suitable Python LocalObject, throw INV_OBJREF.
+  //
+  // Caller must hold the Python interpreter lock.
+  static PyObject* getPyObjectForLocalObject(CORBA::LocalObject_ptr lobj);
+
+
   ////////////////////////////////////////////////////////////////////////////
   // PyUserException is a special CORBA::UserException                      //
   ////////////////////////////////////////////////////////////////////////////
@@ -781,22 +1021,46 @@ public:
 				   // this object is deleted.
   };
 
+
   ////////////////////////////////////////////////////////////////////////////
   // InterpreterUnlocker releases the Python interpreter lock               //
   ////////////////////////////////////////////////////////////////////////////
 
   class InterpreterUnlocker {
   public:
-    InterpreterUnlocker() {
+    inline InterpreterUnlocker() {
       tstate_ = PyEval_SaveThread();
     }
-    ~InterpreterUnlocker() {
+    inline ~InterpreterUnlocker() {
       PyEval_RestoreThread(tstate_);
     }
+    inline void lock() {
+      PyEval_RestoreThread(tstate_);
+    }
+    inline void unlock() {
+      tstate_ = PyEval_SaveThread();
+    }
   private:
     PyThreadState* tstate_;
   };
 
+  ////////////////////////////////////////////////////////////////////////////
+  // ValueTrackerClearer safely clears a ValueTracker                       //
+  ////////////////////////////////////////////////////////////////////////////
+
+  class ValueTrackerClearer {
+  public:
+    inline ValueTrackerClearer(cdrStream& s) : s_(s) {}
+    inline ~ValueTrackerClearer() {
+      if (s_.valueTracker()) {
+        InterpreterUnlocker u;
+        s_.clearValueTracker();
+      }
+    };
+  private:
+    cdrStream& s_;
+  };
+
 
   ////////////////////////////////////////////////////////////////////////////
   // PyUnlockingCdrStream unlocks the interpreter lock around blocking calls//
