@@ -28,9 +28,33 @@
 //    Versions of ORB object ref functions which deal with Python
 //    objects, rather than C++ objects
 
-// $Id: pyObjectRef.cc,v 1.1.2.19 2003/07/28 15:44:21 dgrisby Exp $
-
+// $Id: pyObjectRef.cc,v 1.1.4.7 2006/07/26 17:50:43 dgrisby Exp $
 // $Log: pyObjectRef.cc,v $
+// Revision 1.1.4.7  2006/07/26 17:50:43  dgrisby
+// Reuse existing omniIOR object when converting C++ object reference to Python.
+//
+// Revision 1.1.4.6  2006/07/19 09:40:39  dgrisby
+// Track ORB core changes.
+//
+// Revision 1.1.4.5  2006/05/15 10:26:11  dgrisby
+// More relaxation of requirements for old-style classes, for Python 2.5.
+//
+// Revision 1.1.4.4  2005/06/24 17:36:01  dgrisby
+// Support for receiving valuetypes inside Anys; relax requirement for
+// old style classes in a lot of places.
+//
+// Revision 1.1.4.3  2005/04/25 18:27:41  dgrisby
+// Maintain forwarded location when narrowing forwarded references.
+//
+// Revision 1.1.4.2  2005/01/07 00:22:32  dgrisby
+// Big merge from omnipy2_develop.
+//
+// Revision 1.1.2.21  2004/04/05 09:06:42  dgrisby
+// Bidirectional servers didn't work.
+//
+// Revision 1.1.2.20  2004/03/02 15:33:57  dgrisby
+// Support persistent server id.
+//
 // Revision 1.1.2.19  2003/07/28 15:44:21  dgrisby
 // Unlock interpreter during string_to_object.
 //
@@ -102,6 +126,10 @@
 #include <inProcessIdentity.h>
 #include <objectAdapter.h>
 #include <omniORB4/omniURI.h>
+#include <giopStrand.h>
+#include <giopStream.h>
+#include <omniCurrent.h>
+#include <poaimpl.h>
 
 OMNI_USING_NAMESPACE(omni)
 
@@ -186,7 +214,7 @@ omniPy::createPyCorbaObjRef(const char*             targetRepoId,
       PyObject* targetClass = PyDict_GetItemString(pyomniORBobjrefMap,
 						   (char*)targetRepoId);
 
-      if (!PyClass_IsSubclass(objrefClass, targetClass)) {
+      if (!omniPy::isSubclass(objrefClass, targetClass)) {
 	// Actual type is not derived from the target. Surprisingly
 	// enough, this is valid -- the repoId in an object reference
 	// is not necessarily that of the most derived type for the
@@ -225,8 +253,7 @@ omniPy::createPyCorbaObjRef(const char*             targetRepoId,
 
   if (fullTypeUnknown) {
     PyObject* idstr = PyString_FromString(actualRepoId);
-    PyDict_SetItemString(((PyInstanceObject*)pyobjref)->in_dict,
-			 (char*)"_NP_RepositoryId", idstr);
+    PyObject_SetAttrString(pyobjref, (char*)"_NP_RepositoryId", idstr);
     Py_DECREF(idstr);
   }
   omniPy::setTwin(pyobjref, (CORBA::Object_ptr)objref, OBJREF_TWIN);
@@ -303,7 +330,8 @@ omniPy::createObjRef(const char*    	targetRepoId,
 		     omniIOR*       	ior,
 		     CORBA::Boolean 	locked,
 		     omniIdentity*  	id,
-		     CORBA::Boolean     type_verified)
+		     CORBA::Boolean     type_verified,
+		     CORBA::Boolean     is_forwarded)
 {
   ASSERT_OMNI_TRACEDMUTEX_HELD(*omni::internalLock, locked);
   OMNIORB_ASSERT(targetRepoId);
@@ -338,20 +366,56 @@ omniPy::createObjRef(const char*    	targetRepoId,
     objref->pd_flags.type_verified = 0;
   }
 
+  if (is_forwarded) {
+    omniORB::logs(10, "Reference has been forwarded.");
+    objref->pd_flags.forward_location = 1;
+  }
+
   {
     omni_optional_lock sync(*omni::internalLock, locked, locked);
     id->gainRef(objref);
   }
 
+  if (orbParameters::persistentId.length()) {
+    // Check to see if we need to re-write the IOR.
+
+    omniIOR::IORExtraInfoList& extra = ior->getIORInfo()->extraInfo();
+
+    for (CORBA::ULong index = 0; index < extra.length(); index++) {
+
+      if (extra[index]->compid == IOP::TAG_OMNIORB_PERSISTENT_ID)
+
+	if (!id->inThisAddressSpace()) {
+
+	  omniORB::logs(15, "Re-write local persistent object reference.");
+
+	  omniObjRef* new_objref;
+	  omniIORHints hints(0);
+	  {
+	    omni_optional_lock sync(*internalLock, locked, locked);
+
+	    omniIOR* new_ior = new omniIOR(ior->repositoryID(),
+					   id->key(), id->keysize(), hints);
+
+	    new_objref = createObjRef(targetRepoId, new_ior,
+				      1, 0, type_verified);
+	  }
+	  releaseObjRef(objref);
+	  objref = new_objref;
+	}
+      break;
+    }
+  }
   return objref;
 }
 
 
 omniObjRef*
-omniPy::createLocalObjRef(const char*        mostDerivedRepoId,
-			  const char*        targetRepoId,
-			  omniObjTableEntry* entry,
-			  CORBA::Boolean     type_verified)
+omniPy::createLocalObjRef(const char*         mostDerivedRepoId,
+			  const char*         targetRepoId,
+			  omniObjTableEntry*  entry,
+			  omniObjRef*         orig_ref,
+			  CORBA::Boolean      type_verified)
 {
   ASSERT_OMNI_TRACEDMUTEX_HELD(*omni::internalLock, 1);
   OMNIORB_ASSERT(targetRepoId);
@@ -391,17 +455,17 @@ omniPy::createLocalObjRef(const char*        mostDerivedRepoId,
     }
   }
   // Reach here if we have to create a new objref.
-  omniIOR* ior = new omniIOR(mostDerivedRepoId,
-			     entry->key(), entry->keysize());
-
+  omniIOR* ior = orig_ref->_getIOR();
   return omniPy::createObjRef(targetRepoId, ior, 1, entry, type_verified);
 }
 
 omniObjRef*
-omniPy::createLocalObjRef(const char* mostDerivedRepoId,
-			  const char* targetRepoId,
-			  const _CORBA_Octet* key, int keysize,
-			  CORBA::Boolean type_verified)
+omniPy::createLocalObjRef(const char* 	      mostDerivedRepoId,
+			  const char* 	      targetRepoId,
+			  const _CORBA_Octet* key,
+			  int                 keysize,
+			  omniObjRef*         orig_ref,
+			  CORBA::Boolean      type_verified)
 {
   ASSERT_OMNI_TRACEDMUTEX_HELD(*omni::internalLock, 1);
   OMNIORB_ASSERT(targetRepoId);
@@ -415,11 +479,10 @@ omniPy::createLocalObjRef(const char* mostDerivedRepoId,
 
   if (entry)
     return createLocalObjRef(mostDerivedRepoId, targetRepoId,
-			     entry, type_verified);
+			     entry, orig_ref, type_verified);
 
-  omniIOR* ior = new omniIOR(mostDerivedRepoId, key, keysize);
-
-  return createObjRef(targetRepoId,ior,1,entry,type_verified);
+  omniIOR* ior = orig_ref->_getIOR();
+  return createObjRef(targetRepoId,ior,1,0,type_verified);
 }
 
 
@@ -433,7 +496,6 @@ omniPy::makeLocalObjRef(const char* targetRepoId,
 
   omniObjRef* ooref = objref->_PR_getobj();
   omniObjRef* newooref;
-  omniObjKey  key;
 
   {
     omni_tracedmutex_lock sync(*omni::internalLock);
@@ -442,12 +504,13 @@ omniPy::makeLocalObjRef(const char* targetRepoId,
 
     if (entry)
       newooref = omniPy::createLocalObjRef(ooref->_mostDerivedRepoId(),
-					   targetRepoId, entry, 1);
+					   targetRepoId, entry, ooref, 1);
     else
       newooref = omniPy::createLocalObjRef(ooref->_mostDerivedRepoId(),
 					   targetRepoId,
 					   ooref->_identity()->key(),
-					   ooref->_identity()->keysize(), 1);
+					   ooref->_identity()->keysize(),
+					   ooref, 1);
   }
   return (CORBA::Object_ptr)newooref->_ptrToObjRef(CORBA::Object::_PD_repoId);
 }
@@ -462,10 +525,6 @@ omniPy::copyObjRefArgument(PyObject* pytargetRepoId, PyObject* pyobjref,
     Py_INCREF(Py_None);
     return Py_None;
   }
-  if (!PyInstance_Check(pyobjref)) {
-    // Not an objref
-    OMNIORB_THROW(BAD_PARAM, BAD_PARAM_WrongPythonType, compstatus);
-  }
   CORBA::Object_ptr objref = (CORBA::Object_ptr)getTwin(pyobjref, OBJREF_TWIN);
   if (!objref) {
     // Not an objref
@@ -496,15 +555,13 @@ omniPy::copyObjRefArgument(PyObject* pytargetRepoId, PyObject* pyobjref,
   // so, we can just incref the existing Python objref and return it;
   // if not, we have to build a new C++ objref.
 
-  if (!PyDict_GetItemString(((PyInstanceObject*)pyobjref)->in_dict,
-			    (char*)"_NP_RepositoryId")) {
+  if (!PyObject_HasAttrString(pyobjref, (char*)"_NP_RepositoryId")) {
 
     PyObject* targetClass = PyDict_GetItem(pyomniORBobjrefMap,
 					   pytargetRepoId);
     OMNIORB_ASSERT(targetClass);
 
-    if (PyClass_IsSubclass((PyObject*)((PyInstanceObject*)pyobjref)->in_class,
-			   targetClass)) {
+    if (omniPy::isInstance(pyobjref, targetClass)) {
       Py_INCREF(pyobjref);
       return pyobjref;
     }
@@ -588,6 +645,22 @@ omniPy::UnMarshalObjRef(const char* repoId, cdrStream& s)
     // this has been accepted as a valid behaviour in GIOP 1.1/IIOP 1.1.
     // 
     omniIOR* ior = new omniIOR(id._retn(),profiles._retn());
+
+    giopStream* gs = giopStream::downcast(&s);
+    if (gs) {
+      giopStrand& g = (giopStrand&)*gs;
+      if (g.biDir && !g.isClient()) {
+	// Check the POA policy to see if the servant's POA is willing
+	// to use bidirectional on its callback objects.
+	omniCurrent* current = omniCurrent::get();
+	omniCallDescriptor* desc = ((current)? current->callDescriptor() :0);
+
+	if (desc && desc->poa() && desc->poa()->acceptBiDirectional()) {
+	  const char* sendfrom = g.connection->peeraddress();
+	  omniIOR::add_TAG_OMNIORB_BIDIR(sendfrom,*ior);
+	}
+      }
+    }
     omniObjRef* objref = omniPy::createObjRef(repoId,ior,0);
 
     if (!objref) OMNIORB_THROW(MARSHAL, MARSHAL_InvalidIOR,
