@@ -27,14 +27,29 @@
 // Description:
 //    omniORB API functions
 
-// $Id: pyomniFunc.cc,v 1.1.2.13 2003/10/28 13:59:44 dgrisby Exp $
-
+// $Id: pyomniFunc.cc,v 1.1.4.7 2006/01/17 17:38:21 dgrisby Exp $
 // $Log: pyomniFunc.cc,v $
-// Revision 1.1.2.13  2003/10/28 13:59:44  dgrisby
-// myIPAddresses function.
+// Revision 1.1.4.7  2006/01/17 17:38:21  dgrisby
+// Expose omniORB.setClientConnectTimeout function.
 //
-// Revision 1.1.2.12  2003/09/01 09:53:39  dgrisby
-// Avoid deadlock in dummy thread deletion.
+// Revision 1.1.4.6  2005/11/09 12:33:32  dgrisby
+// Support POA LocalObjects.
+//
+// Revision 1.1.4.5  2005/07/22 17:41:07  dgrisby
+// Update from omnipy2_develop.
+//
+// Revision 1.1.4.4  2005/06/24 17:36:00  dgrisby
+// Support for receiving valuetypes inside Anys; relax requirement for
+// old style classes in a lot of places.
+//
+// Revision 1.1.4.3  2005/04/14 13:50:59  dgrisby
+// New traceTime, traceInvocationReturns functions; removal of omniORB::logf.
+//
+// Revision 1.1.4.2  2005/01/07 00:22:33  dgrisby
+// Big merge from omnipy2_develop.
+//
+// Revision 1.1.4.1  2003/03/23 21:51:56  dgrisby
+// New omnipy3_develop branch.
 //
 // Revision 1.1.2.11  2002/11/27 00:18:25  dgrisby
 // Per thread / per objref timeouts.
@@ -107,7 +122,7 @@ static CORBA::Boolean transientEH(void* cookie, CORBA::ULong retries,
     
     if (!r) {
       if (omniORB::trace(1)) {
-	omniORB::logf("Python TRANSIENT exception handler failed. "
+	omniORB::logs(1, "Python TRANSIENT exception handler failed. "
 		      "Traceback follows:");
 	PyErr_Print();
       }
@@ -116,9 +131,8 @@ static CORBA::Boolean transientEH(void* cookie, CORBA::ULong retries,
       return 0;
     }
     if (!PyInt_Check(r)) {
-      if (omniORB::trace(1))
-	omniORB::logf("Python TRANSIENT exception handler returned an "
-		      "invalid object.");
+      omniORB::logs(1, "Python TRANSIENT exception handler returned an "
+		    "invalid object.");
       Py_DECREF(r);
       return 0;
     }
@@ -153,7 +167,7 @@ static CORBA::Boolean commFailureEH(void* cookie, CORBA::ULong retries,
     
     if (!r) {
       if (omniORB::trace(1)) {
-	omniORB::logf("Python COMM_FAILURE exception handler failed. "
+	omniORB::logs(1, "Python COMM_FAILURE exception handler failed. "
 		      "Traceback follows:");
 	PyErr_Print();
       }
@@ -162,9 +176,8 @@ static CORBA::Boolean commFailureEH(void* cookie, CORBA::ULong retries,
       return 0;
     }
     if (!PyInt_Check(r)) {
-      if (omniORB::trace(1))
-	omniORB::logf("Python COMM_FAILURE exception handler returned an "
-		      "invalid object.");
+      omniORB::logs(1, "Python COMM_FAILURE exception handler returned an "
+		    "invalid object.");
       Py_DECREF(r);
       return 0;
     }
@@ -198,7 +211,7 @@ static CORBA::Boolean systemEH(void* cookie, CORBA::ULong retries,
     
     if (!r) {
       if (omniORB::trace(1)) {
-	omniORB::logf("Python SystemException handler failed. "
+	omniORB::logs(1, "Python SystemException handler failed. "
 		      "Traceback follows:");
 	PyErr_Print();
       }
@@ -207,9 +220,8 @@ static CORBA::Boolean systemEH(void* cookie, CORBA::ULong retries,
       return 0;
     }
     if (!PyInt_Check(r)) {
-      if (omniORB::trace(1))
-	omniORB::logf("Python SystemException handler returned an "
-		      "invalid object.");
+      omniORB::logs(1, "Python SystemException handler returned an "
+		    "invalid object.");
       Py_DECREF(r);
       return 0;
     }
@@ -243,14 +255,19 @@ omniPy::ensureOmniThread()
 
   // Get the threading.Thread object for this thread.
   PyObject* threading = PyImport_ImportModule((char*)"threading");
-  if (!threading)
+  if (!threading) {
+    omniORB::logs(1, "Unable to import Python threading module.");
     return 0;
+  }
 
   PyObject* current = PyObject_CallMethod(threading,
 					  (char*)"currentThread",
 					  (char*)"");
-  if (!current)
+  if (!current) {
+    omniORB::logs(1, "Unexpected exception calling threading.currentThread.");
+    if (omniORB::trace(1)) PyErr_Print();
     return 0;
+  }
 
   // Create a dummy omni_thread
   self = omni_thread::create_dummy();
@@ -265,8 +282,10 @@ omniPy::ensureOmniThread()
   PyObject* hook = PyObject_CallMethod(omniPy::pyomniORBmodule,
 				       (char*)"omniThreadHook", (char*)"O",
 				       current);
-  if (!hook)
-    PyErr_Print();
+  if (!hook) {
+    omniORB::logs(1, "Unexpected exception calling omniThreadHook.");
+    if (omniORB::trace(1)) PyErr_Print();
+  }
 
   Py_XDECREF(hook);
   Py_DECREF(cobj);
@@ -303,8 +322,6 @@ extern "C" {
 
 
     RAISE_PY_BAD_PARAM_IF(!PyCallable_Check(pyfn), BAD_PARAM_WrongPythonType);
-    RAISE_PY_BAD_PARAM_IF(pyobjref && !PyInstance_Check(pyobjref),
-			  BAD_PARAM_WrongPythonType);
 
     if (pyobjref) {
       CORBA::Object_ptr objref =
@@ -351,8 +368,6 @@ extern "C" {
       return 0;
 
     RAISE_PY_BAD_PARAM_IF(!PyCallable_Check(pyfn), BAD_PARAM_WrongPythonType);
-    RAISE_PY_BAD_PARAM_IF(pyobjref && !PyInstance_Check(pyobjref),
-			  BAD_PARAM_WrongPythonType);
 
     if (pyobjref) {
       CORBA::Object_ptr objref =
@@ -400,8 +415,6 @@ extern "C" {
       return 0;
 
     RAISE_PY_BAD_PARAM_IF(!PyCallable_Check(pyfn), BAD_PARAM_WrongPythonType);
-    RAISE_PY_BAD_PARAM_IF(pyobjref && !PyInstance_Check(pyobjref),
-			  BAD_PARAM_WrongPythonType);
 
     if (pyobjref) {
       CORBA::Object_ptr objref =
@@ -450,11 +463,36 @@ extern "C" {
     return 0;
   }
 
+  static char traceExceptions_doc [] =
+  "traceExceptions(int) -> None\n"
+  "traceExceptions()    -> int\n"
+  "\n"
+  "Set or get the omniORB exception tracing flag.\n";
+
+  static PyObject* pyomni_traceExceptions(PyObject* self, PyObject* args)
+  {
+    if (PyTuple_GET_SIZE(args) == 0) {
+      return PyInt_FromLong(omniORB::traceExceptions);
+    }
+    else if (PyTuple_GET_SIZE(args) == 1) {
+      PyObject* pytl = PyTuple_GET_ITEM(args, 0);
+
+      if (PyInt_Check(pytl)) {
+	omniORB::traceExceptions = PyInt_AS_LONG(pytl);
+	Py_INCREF(Py_None);
+	return Py_None;
+      }
+    }
+    PyErr_SetString(PyExc_TypeError,
+		    (char*)"Operation requires a single integer argument");
+    return 0;
+  }
+
   static char traceInvocations_doc [] =
   "traceInvocations(int) -> None\n"
   "traceInvocations()    -> int\n"
   "\n"
-  "Set or get the omniORB invocaton tracing flag.\n";
+  "Set or get the omniORB invocation tracing flag.\n";
 
   static PyObject* pyomni_traceInvocations(PyObject* self, PyObject* args)
   {
@@ -475,6 +513,31 @@ extern "C" {
     return 0;
   }
 
+  static char traceInvocationReturns_doc [] =
+  "traceInvocationReturns(int) -> None\n"
+  "traceInvocationReturns()    -> int\n"
+  "\n"
+  "Set or get the omniORB invocation return tracing flag.\n";
+
+  static PyObject* pyomni_traceInvocationReturns(PyObject* self, PyObject* args)
+  {
+    if (PyTuple_GET_SIZE(args) == 0) {
+      return PyInt_FromLong(omniORB::traceInvocationReturns);
+    }
+    else if (PyTuple_GET_SIZE(args) == 1) {
+      PyObject* pytl = PyTuple_GET_ITEM(args, 0);
+
+      if (PyInt_Check(pytl)) {
+	omniORB::traceInvocationReturns = PyInt_AS_LONG(pytl);
+	Py_INCREF(Py_None);
+	return Py_None;
+      }
+    }
+    PyErr_SetString(PyExc_TypeError,
+		    (char*)"Operation requires a single integer argument");
+    return 0;
+  }
+
   static char traceThreadId_doc [] =
   "traceThreadId(int) -> None\n"
   "traceThreadId()    -> int\n"
@@ -500,6 +563,31 @@ extern "C" {
     return 0;
   }
 
+  static char traceTime_doc [] =
+  "traceTime(int) -> None\n"
+  "traceTime()    -> int\n"
+  "\n"
+  "Set or get the omniORB time tracing flag.\n";
+
+  static PyObject* pyomni_traceTime(PyObject* self, PyObject* args)
+  {
+    if (PyTuple_GET_SIZE(args) == 0) {
+      return PyInt_FromLong(omniORB::traceTime);
+    }
+    else if (PyTuple_GET_SIZE(args) == 1) {
+      PyObject* pytl = PyTuple_GET_ITEM(args, 0);
+
+      if (PyInt_Check(pytl)) {
+	omniORB::traceTime = PyInt_AS_LONG(pytl);
+	Py_INCREF(Py_None);
+	return Py_None;
+      }
+    }
+    PyErr_SetString(PyExc_TypeError,
+		    (char*)"Operation requires a single integer argument");
+    return 0;
+  }
+
 
   static char log_doc [] =
   "log(level, string)\n"
@@ -596,16 +684,20 @@ extern "C" {
   static PyObject* pyomni_minorCodeToString(PyObject* self, PyObject* args)
   {
     PyObject* pyexc;
-    PyObject* pyrepoId = 0;
-    PyObject* pyminor = 0;
+    PyObject* pyrepoId;
+    PyObject* pyminor;
 
     if (!PyArg_ParseTuple(args, (char*)"O", &pyexc))
       return 0;
 
-    if (PyInstance_Check(pyexc)) {
-      pyrepoId = PyObject_GetAttrString(pyexc, (char*)"_NP_RepositoryId");
-      pyminor  = PyObject_GetAttrString(pyexc, (char*)"minor");
-    }
+    pyrepoId = PyObject_GetAttrString(pyexc, (char*)"_NP_RepositoryId");
+    if (!pyrepoId)
+      PyErr_Clear();
+
+    pyminor  = PyObject_GetAttrString(pyexc, (char*)"minor");
+    if (!pyminor)
+      PyErr_Clear();
+
     if (!(pyrepoId && PyString_Check(pyrepoId) &&
 	  pyminor  && (PyInt_Check(pyminor) || PyLong_Check(pyminor)))) {
       Py_INCREF(Py_None);
@@ -616,12 +708,17 @@ extern "C" {
     CORBA::ULong minor;
     if (PyInt_Check(pyminor))
       minor = PyInt_AS_LONG(pyminor);
-    else
+    else {
       minor = PyLong_AsUnsignedLong(pyminor);
+      if (minor == (CORBA::ULong)-1 && PyErr_Occurred())
+	PyErr_Clear();
+    }
 
     const char* str = 0;
 
     if (0) {
+      // Empty case to allow us to have a big chain of else ifs due to
+      // the macro expansion below.
     }
 #define ToStringIfMatch(name) \
     else if (!strcmp(repoId, "IDL:omg.org/CORBA/" #name ":1.0")) \
@@ -658,8 +755,7 @@ extern "C" {
       PyObject* pyobjref;
       if (!PyArg_ParseTuple(args, (char*)"Oi", &pyobjref, &timeout))
 	return 0;
-      RAISE_PY_BAD_PARAM_IF(!PyInstance_Check(pyobjref),
-			    BAD_PARAM_WrongPythonType);
+
       CORBA::Object_ptr objref =
 	(CORBA::Object_ptr)omniPy::getTwin(pyobjref, OBJREF_TWIN);
 
@@ -683,8 +779,29 @@ extern "C" {
     if (!PyArg_ParseTuple(args, (char*)"i", &timeout))
       return 0;
 
-    omniPy::ensureOmniThread();
-    omniORB::setClientThreadCallTimeout(timeout);
+    try {
+      omniPy::ensureOmniThread();
+      omniORB::setClientThreadCallTimeout(timeout);
+    }
+    OMNIPY_CATCH_AND_HANDLE_SYSTEM_EXCEPTIONS
+
+    Py_INCREF(Py_None);
+    return Py_None;
+  }
+
+  static char setClientConnectTimeout_doc [] =
+  "setClientConnectTimeout(millisecs)\n"
+  "\n"
+  "Set the client connection timeout.\n";
+
+  static PyObject* pyomni_setClientConnectTimeout(PyObject* self,
+						  PyObject* args)
+  {
+    int timeout;
+    if (!PyArg_ParseTuple(args, (char*)"i", &timeout))
+      return 0;
+
+    omniORB::setClientConnectTimeout(timeout);
 
     Py_INCREF(Py_None);
     return Py_None;
@@ -716,6 +833,61 @@ extern "C" {
     return pyaddrs;
   }
 
+  static char setPersistentServerIdentifier_doc [] =
+  "setPersistentServerIdentifier(ident)\n"
+  "\n"
+  "Sets an octet sequence used to persistently identify \"this\"\n"
+  "server. Stored object references matching this identifier are\n"
+  "re-written to use the current endpoint details.\n";
+
+  static PyObject* pyomni_setPersistentServerIdentifier(PyObject* self,
+							PyObject* args)
+  {
+    char* idstr;
+    int   idlen;
+
+    if (!PyArg_ParseTuple(args, (char*)"s#", &idstr, &idlen))
+      return 0;
+
+    CORBA::OctetSeq idseq(idlen, idlen, (CORBA::Octet*)idstr, 0);
+    try {
+      omniPy::InterpreterUnlocker _u;
+      omniORB::setPersistentServerIdentifier(idseq);
+    }
+    OMNIPY_CATCH_AND_HANDLE_SYSTEM_EXCEPTIONS
+
+    Py_INCREF(Py_None);
+    return Py_None;
+  }
+
+  static char locationForward_doc [] =
+  "locationForward(object_reference, new_location)\n"
+  "\n"
+  "Reconfigures the object reference to refer to the new location,\n"
+  "as if a GIOP location forward message had been received.\n";
+
+  static PyObject* pyomni_locationForward(PyObject* self, PyObject* args)
+  {
+    PyObject* pyold;
+    PyObject* pynew;
+    if (!PyArg_ParseTuple(args, (char*)"OO", &pyold, &pynew))
+      return 0;
+
+    CORBA::Object_ptr oldobj =
+      (CORBA::Object_ptr)omniPy::getTwin(pyold, OBJREF_TWIN);
+
+    RAISE_PY_BAD_PARAM_IF(!oldobj, BAD_PARAM_WrongPythonType);
+
+    CORBA::Object_ptr newobj =
+      (CORBA::Object_ptr)omniPy::getTwin(pynew, OBJREF_TWIN);
+
+    RAISE_PY_BAD_PARAM_IF(!newobj, BAD_PARAM_WrongPythonType);
+
+    omni::locationForward(oldobj->_PR_getobj(), newobj->_PR_getobj(), 0);
+
+    Py_INCREF(Py_None);
+    return Py_None;
+  }
 
   static PyMethodDef pyomni_methods[] = {
     {(char*)"installTransientExceptionHandler",
@@ -734,14 +906,26 @@ extern "C" {
      pyomni_traceLevel,
      METH_VARARGS, traceLevel_doc},
 
+    {(char*)"traceExceptions",
+     pyomni_traceExceptions,
+     METH_VARARGS, traceExceptions_doc},
+
     {(char*)"traceInvocations",
      pyomni_traceInvocations,
      METH_VARARGS, traceInvocations_doc},
 
+    {(char*)"traceInvocationReturns",
+     pyomni_traceInvocationReturns,
+     METH_VARARGS, traceInvocationReturns_doc},
+
     {(char*)"traceThreadId",
      pyomni_traceThreadId,
      METH_VARARGS, traceThreadId_doc},
 
+    {(char*)"traceTime",
+     pyomni_traceTime,
+     METH_VARARGS, traceTime_doc},
+
     {(char*)"log",
      pyomni_log,
      METH_VARARGS, log_doc},
@@ -766,10 +950,22 @@ extern "C" {
      pyomni_setClientThreadCallTimeout,
      METH_VARARGS, setClientThreadCallTimeout_doc},
 
+    {(char*)"setClientConnectTimeout",
+     pyomni_setClientConnectTimeout,
+     METH_VARARGS, setClientConnectTimeout_doc},
+
     {(char*)"myIPAddresses",
      pyomni_myIPAddresses,
      METH_VARARGS, myIPAddresses_doc},
 
+    {(char*)"setPersistentServerIdentifier",
+     pyomni_setPersistentServerIdentifier,
+     METH_VARARGS, setPersistentServerIdentifier_doc},
+
+    {(char*)"locationForward",
+     pyomni_locationForward,
+     METH_VARARGS, locationForward_doc},
+
     {NULL,NULL}
   };
 }
