@@ -26,16 +26,50 @@
 #   
 #   Back-end for Python
 
-# $Id: python.py,v 1.29.2.17 2003/11/19 16:56:36 dgrisby Exp $
+# $Id: python.py,v 1.33.2.14 2008/02/01 16:29:17 dgrisby Exp $
 # $Log: python.py,v $
-# Revision 1.29.2.17  2003/11/19 16:56:36  dgrisby
-# Extern package option to omniidl backend. Thanks Rene Jager.
+# Revision 1.33.2.14  2008/02/01 16:29:17  dgrisby
+# Error with implementation of operations with names clashing with
+# Python keywords.
 #
-# Revision 1.29.2.16  2003/04/25 15:45:16  dgrisby
-# Fix bug with enum members that clash with keywords.
+# Revision 1.33.2.13  2006/10/11 17:44:14  dgrisby
+# None is not a keyword, but it cannot be assigned to.
 #
-# Revision 1.29.2.15  2003/04/16 13:07:59  dgrisby
-# Option to generate example implementations.
+# Revision 1.33.2.12  2006/09/29 16:48:03  dgrisby
+# Stub changes broke use of package prefix. Thanks Teemu Torma.
+#
+# Revision 1.33.2.11  2006/09/07 15:28:57  dgrisby
+# Remove obsolete check for presence of omniORB.StructBase.
+#
+# Revision 1.33.2.10  2006/06/21 14:46:26  dgrisby
+# Invalid generated code for structs nested inside valuetypes.
+#
+# Revision 1.33.2.9  2006/01/19 17:28:44  dgrisby
+# Merge from omnipy2_develop.
+#
+# Revision 1.33.2.8  2006/01/18 19:25:13  dgrisby
+# Bug inheriting a valuetype from a typedef.
+#
+# Revision 1.33.2.7  2005/07/29 11:21:36  dgrisby
+# Fix long-standing problem with module re-opening by #included files.
+#
+# Revision 1.33.2.6  2005/01/07 00:22:34  dgrisby
+# Big merge from omnipy2_develop.
+#
+# Revision 1.33.2.5  2004/03/24 22:28:50  dgrisby
+# TypeCodes / truncation for inherited state members were broken.
+#
+# Revision 1.33.2.4  2004/02/16 10:14:18  dgrisby
+# Use stream based copy for local calls.
+#
+# Revision 1.33.2.3  2003/07/10 22:13:25  dgrisby
+# Abstract interface support.
+#
+# Revision 1.33.2.2  2003/05/20 17:10:24  dgrisby
+# Preliminary valuetype support.
+#
+# Revision 1.33.2.1  2003/03/23 21:51:56  dgrisby
+# New omnipy3_develop branch.
 #
 # Revision 1.29.2.14  2002/11/25 21:31:09  dgrisby
 # Friendly error messages with file errors, remove code to kill POA
@@ -190,6 +224,7 @@ cpp_args = ["-D__OMNIIDL_PYTHON__"]
 usage_string = """\
   -Wbstdout       Send generated stubs to stdout rather than a file
   -Wbinline       Output stubs for #included files in line with the main file
+  -Wbfactories    Register value factories for all valuetypes
   -Wbpackage=p    Put both Python modules and stub files in package p
   -Wbmodules=p    Put Python modules in package p
   -Wbstubs=p      Put stub files in package p
@@ -205,6 +240,9 @@ pymodule_template = """\
 #
 # Python module @module@ generated by omniidl
 
+import omniORB
+omniORB.updateModule("@package@@module@")
+
 # ** 1. Stub files contributing to this module
 
 # ** 2. Sub-modules
@@ -218,7 +256,7 @@ import omniORB, _omnipy
 from omniORB import CORBA, PortableServer
 _0_CORBA = CORBA
 
-_omnipy.checkVersion(2,0, __file__)
+_omnipy.checkVersion(3,0, __file__)
 """
 
 file_end = """\
@@ -252,15 +290,15 @@ _0_@s_imodname@ = omniORB.openModule("@package@@s_imodname@")"""
 
 forward_interface = """\
 
-# interface @ifid@;
-_0_@modname@._d_@ifid@ = (omniORB.tcInternal.tv_objref, "@repoId@", "@ifid@")
+# @abstract@interface @ifid@;
+_0_@modname@._d_@ifid@ = (omniORB.tcInternal.@tvkind@, "@repoId@", "@ifid@")
 omniORB.typeMapping["@repoId@"] = _0_@modname@._d_@ifid@"""
 
 
 interface_class = """\
 
-# interface @ifid@
-_0_@modname@._d_@ifid@ = (omniORB.tcInternal.tv_objref, "@repoId@", "@ifid@")
+# @abstract@interface @ifid@
+_0_@modname@._d_@ifid@ = (omniORB.tcInternal.@tvkind@, "@repoId@", "@ifid@")
 omniORB.typeMapping["@repoId@"] = _0_@modname@._d_@ifid@
 _0_@modname@.@ifid@ = omniORB.newEmptyClass()
 class @ifid@ @inherits@:
@@ -287,12 +325,12 @@ attribute_set_descriptor = """\
 @ifid@._d__set_@attr@ = ((@descr@,),(),None)"""
 
 operation_descriptor = """\
-@ifid@._d_@opname@ = (@inds@, @outds@, @excs@@ctxts@)"""
+@ifid@._d_@opname@ = (@inds@, @outds@, @excs@@options@)"""
 
 objref_class = """\
 
 # @ifid@ object reference
-class _objref_@ifid@ @inherits@:
+class _objref_@ifid@ (@inherits@):
     _NP_RepositoryId = @ifid@._NP_RepositoryId
 
     def __init__(self):"""
@@ -311,7 +349,7 @@ objref_attribute_set = """
         return _omnipy.invoke(self, "_set_@attr@", _0_@modname@.@ifid@._d__set_@attr@, args)"""
 objref_operation = """
     def @opname@(self, *args):
-        return _omnipy.invoke(self, "@opname@", _0_@modname@.@ifid@._d_@opname@, args)"""
+        return _omnipy.invoke(self, "@r_opname@", _0_@modname@.@ifid@._d_@opname@, args)"""
 objref_methods = """
     __methods__ = @methods@"""
 
@@ -336,10 +374,15 @@ skeleton_inheritmap = """\
 skeleton_end = """
 @ifid@._omni_skeleton = @ifid@
 _0_@s_modname@.@ifid@ = @ifid@
+omniORB.registerSkeleton(@ifid@._NP_RepositoryId, @ifid@)
 del @ifid@
 __name__ = "@package@@modname@"\
 """
 
+skeleton_set_skel = """
+@ifid@._omni_skeleton = @ifid@
+"""
+
 constant_at_module_scope = """\
 _0_@modname@.@cname@ = @value@"""
 
@@ -403,10 +446,14 @@ omniORB.typeMapping["@repoId@"] = _d_@sname@"""
 struct_class = """
 # struct @sname@
 _0_@scopedname@ = omniORB.newEmptyClass()
-class @sname@:
+class @sname@ (omniORB.StructBase):
     _NP_RepositoryId = "@repoId@"
 """
 
+struct_class_name = """\
+    _NP_ClassName = "@cname@"
+"""
+
 struct_class_init = """\
     def __init__(self@mnames@):"""
 
@@ -493,6 +540,10 @@ class @uname@ (omniORB.Union):
     _NP_RepositoryId = "@repoId@"\
 """
 
+union_class_name = """\
+    _NP_ClassName = "@cname@"
+"""
+
 union_descriptor_at_module_scope = """
 _0_@modname@.@uname@ = @uname@
 
@@ -548,6 +599,74 @@ _tc_@ename@ = omniORB.tcInternal.createTypeCode(_d_@ename@)
 omniORB.registerType(@ename@._NP_RepositoryId, _d_@ename@, _tc_@ename@)"""
 
 
+value_forward_at_module_scope = """\
+# forward valuetype @vname@
+_0_@modname@._d_@vname@ = (omniORB.tcInternal.tv__indirect, ["@repoId@"])
+omniORB.typeMapping["@repoId@"] = _0_@modname@._d_@vname@
+"""
+
+value_class = """
+# valuetype @vname@
+_0_@modname@._d_@vname@ = (omniORB.tcInternal.tv__indirect, ["@repoId@"])
+omniORB.typeMapping["@repoId@"] = _0_@modname@._d_@vname@
+_0_@modname@.@vname@ = omniORB.newEmptyClass()
+
+class @vname@ (@inherits@):
+    _NP_RepositoryId = "@repoId@"
+
+    def __init__(self, *args, **kwargs):
+        if args:
+            if len(args) != @arglen@:
+                raise TypeError("@vname@() takes @arglen@ argument@s@ "
+                                "(%d given)" % len(args))
+            @set_args@
+        if kwargs:
+            self.__dict__.update(kwargs)
+"""
+
+valueabs_class = """\
+class @vname@ (@inherits@):
+    _NP_RepositoryId = "@repoId@"
+
+    def __init__(self, *args, **kwargs):
+        raise RuntimeError("Cannot construct objects of this type.")
+"""
+
+value_register_factory = """\
+omniORB.registerValueFactory(@vname@._NP_RepositoryId, @vname@)
+"""
+
+value_descriptor_at_module_scope = """\
+_0_@modname@.@vname@ = @vname@
+_0_@modname@._d_@vname@  = (omniORB.tcInternal.tv_value, @vname@, @vname@._NP_RepositoryId, "@vname@", @modifier@, @tbaseids@, @basedesc@, @mdescs@)
+_0_@modname@._tc_@vname@ = omniORB.tcInternal.createTypeCode(_0_@modname@._d_@vname@)
+omniORB.registerType(@vname@._NP_RepositoryId, _0_@modname@._d_@vname@, _0_@modname@._tc_@vname@)
+del @vname@
+"""
+
+value_objref_register = """
+omniORB.registerObjref(@ifid@._NP_RepositoryId, _objref_@ifid@)
+_0_@modname@._objref_@ifid@ = _objref_@ifid@
+del _objref_@ifid@"""
+
+
+valuebox = """\
+
+# valuebox @boxname@
+class @boxname@:
+    _NP_RepositoryId = "@repoId@"
+    def __init__(self, *args, **kw):
+        raise RuntimeError("Cannot construct objects of this type.")
+
+_0_@modname@.@boxname@ = @boxname@
+_0_@modname@._d_@boxname@  = (omniORB.tcInternal.tv_value_box, @boxname@, @boxname@._NP_RepositoryId, "@boxname@", @boxdesc@)
+_0_@modname@._tc_@boxname@ = omniORB.tcInternal.createTypeCode(_0_@modname@._d_@boxname@)
+omniORB.registerType(@boxname@._NP_RepositoryId, _0_@modname@._d_@boxname@, _0_@modname@._tc_@boxname@)
+omniORB.registerValueFactory(@boxname@._NP_RepositoryId, @boxname@)
+del @boxname@
+"""
+
+
 example_start = """\
 #!/usr/bin/env python
 
@@ -614,7 +733,6 @@ if __name__ == "__main__":
 """
 
 
-
 # Global state
 imported_files   = {}
 exported_modules = {}
@@ -625,6 +743,7 @@ global_module    = "_GlobalIDL"
 module_package   = ""
 stub_package     = ""
 stub_directory   = ""
+all_factories    = 0
 example_impl     = 0
 extern_stub_pkgs = {}
 
@@ -636,7 +755,7 @@ def error_exit(message):
 def run(tree, args):
     global main_idl_file, imported_files, exported_modules, output_inline
     global global_module, module_package, stub_package, stub_directory
-    global example_impl, extern_stub_pkgs
+    global all_factories, example_impl, extern_stub_pkgs
 
     imported_files.clear()
     exported_modules.clear()
@@ -656,6 +775,9 @@ def run(tree, args):
         elif arg == "inline":
             output_inline = 1
 
+        elif arg == "factories":
+            all_factories = 1
+
         elif arg[:8] == "modules=":
             module_package = arg[8:]
             if module_package != "":
@@ -876,17 +998,18 @@ class PythonVisitor:
     def visitForward(self, node):
         if self.handleImported(node): return
 
-        if node.abstract():
-            sys.stderr.write(node.file() + ":" + str(node.line()) + \
-                             ": omniORBpy does not support abstract " \
-                             "interfaces\n")
-            sys.exit(1)
-
         assert self.at_module_scope
         ifid   = mangle(node.identifier())
         repoId = node.repoId()
-        self.st.out(forward_interface, ifid=ifid,
-                    repoId=repoId, modname=self.modname)
+        if node.abstract():
+            tvkind = "tv_abstract_interface"
+            abstract = "abstract "
+        else:
+            tvkind = "tv_objref"
+            abstract = ""
+
+        self.st.out(forward_interface, ifid=ifid, tvkind=tvkind,
+                    repoId=repoId, abstract=abstract, modname=self.modname)
 
     #
     # Interface
@@ -894,29 +1017,30 @@ class PythonVisitor:
     def visitInterface(self, node):
         if self.handleImported(node): return
 
-        if node.abstract():
-            sys.stderr.write(node.file() + ":" + str(node.line()) + \
-                             ": omniORBpy does not support abstract " \
-                             "interfaces\n")
-            sys.exit(1)
-
         assert self.at_module_scope
         ifid = mangle(node.identifier())
 
         if len(node.inherits()) > 0:
             inheritl = []
             for i in node.inherits():
-                while isinstance(i, idlast.Declarator):
-                    i = i.alias().aliasType()
+                i = i.fullDecl()
                 inheritl.append(dotName(fixupScopedName(i.scopedName())))
             
             inherits = "(" + string.join(inheritl, ", ") + ")"
         else:
             inherits = ""
 
+        if node.abstract():
+            tvkind = "tv_abstract_interface"
+            abstract = "abstract "
+        else:
+            tvkind = "tv_objref"
+            abstract = ""
+
         # Class header
-        self.st.out(interface_class, ifid=ifid, inherits=inherits,
-                    repoId=node.repoId(), modname=self.modname)
+        self.st.out(interface_class, ifid=ifid, tvkind=tvkind,
+                    inherits=inherits, repoId=node.repoId(),
+                    abstract=abstract, modname=self.modname)
 
         # Declarations within the interface
         if len(node.declarations()) > 0:
@@ -957,38 +1081,43 @@ class PythonVisitor:
                                     attr=attr, descr=descr, ifid=ifid)
             else: # Operation
 
-                inds, outds, excs, ctxts = operationToDescriptors(c)
+                inds, outds, excs, ctxts, cv = operationToDescriptors(c)
+
+                options = ""
+
+                if cv:
+                    ctxts = ctxts or "None"
 
                 if ctxts:
-                    ctxts = ", " + ctxts
-                else:
-                    ctxts = ""
+                    options = ", " + ctxts
+
+                if cv:
+                    options = options + ", 1"
 
                 # Output the declaration
                 self.st.out(operation_descriptor,
-                            opname = mangle(c.identifier()),
-                            inds   = inds,
-                            outds  = outds,
-                            excs   = excs,
-                            ctxts  = ctxts,
-                            ifid   = ifid)
+                            opname  = mangle(c.identifier()),
+                            inds    = inds,
+                            outds   = outds,
+                            excs    = excs,
+                            options = options,
+                            ifid    = ifid)
 
         # Objref class
-        if len(node.inherits()) > 0:
+        if node.inherits():
             inheritl = []
             for i in node.inherits():
-                while isinstance(i, idlast.Declarator):
-                    i = i.alias().aliasType()
+                i = i.fullDecl()
                 sn = fixupScopedName(i.scopedName())
                 inheritl.append(dotName(sn[:-1] + ["_objref_" + sn[-1]]))
                 
-            inherits = "(" + string.join(inheritl, ", ") + ")"
+            inherits = string.join(inheritl, ", ")
         else:
-            inherits = "(CORBA.Object)"
+            inherits = "CORBA.Object"
 
         self.st.out(objref_class, ifid=ifid, inherits=inherits)
 
-        if len(node.inherits()) > 0:
+        if node.inherits():
             for inclass in inheritl:
                 self.st.out(objref_inherit_init, inclass=inclass)
         else:
@@ -1022,20 +1151,20 @@ class PythonVisitor:
                 opname = mangle(c.identifier())
                 
                 self.st.out(objref_operation,
-                            opname  = opname,
-                            ifid    = ifid,
-                            modname = self.modname)
+                            opname   = opname,
+                            r_opname = c.identifier(),
+                            ifid     = ifid,
+                            modname  = self.modname)
                 
                 methodl.append('"' + opname + '"')
 
         # __methods__ assignment
         methods = "[" + string.join(methodl, ", ") + "]"
 
-        if len(node.inherits()) > 0:
+        if node.inherits():
             inheritl = []
             for i in node.inherits():
-                while isinstance(i, idlast.Declarator):
-                    i = i.alias().aliasType()
+                i = i.fullDecl()
                 sn = fixupScopedName(i.scopedName())
                 methods = methods + " + " + \
                           dotName(sn[:-1] + ["_objref_" + sn[-1]]) + \
@@ -1049,11 +1178,10 @@ class PythonVisitor:
         self.st.out(objref_register, ifid = ifid, modname = self.modname)
 
         # Skeleton class
-        if len(node.inherits()) > 0:
+        if node.inherits():
             inheritl = []
             for i in node.inherits():
-                while isinstance(i, idlast.Declarator):
-                    i = i.alias().aliasType()
+                i = i.fullDecl()
                 fsn = fixupScopedName(i.scopedName())
                 dsn = dotName(fsn)
                 ssn = skeletonModuleName(dsn)
@@ -1089,16 +1217,17 @@ class PythonVisitor:
                                        ifid + '.' + '_d__set_' + attr)
 
             else: # Operation
-                opname = mangle(c.identifier())
+                opname   = c.identifier()
+                m_opname = mangle(opname)
                 
-                methodl.append('"' + opname + '": ' + '_0_' + self.modname + \
-                               '.' + ifid + '.' + '_d_' + opname)
+                methodl.append('"' + opname + '": ' + '_0_' + self.modname +
+                               '.' + ifid + '.' + '_d_' + m_opname)
 
         methodmap = "{" + string.join(methodl, ", ") + "}"
 
         self.st.out(skeleton_methodmap, methodmap = methodmap)
 
-        if len(node.inherits()) > 0:
+        if node.inherits():
             for inheritclass in inheritl:
                 self.st.out(skeleton_inheritmap, inheritclass = inheritclass)
 
@@ -1249,6 +1378,9 @@ class PythonVisitor:
                     repoId     = node.repoId(),
                     scopedname = dotName(fscopedName))
 
+        if not self.at_module_scope:
+            self.st.out(struct_class_name, cname = dotName(fscopedName))
+
         mnamel = []
         mdescl = []
         for mem in node.members():
@@ -1343,6 +1475,9 @@ class PythonVisitor:
                     repoId = node.repoId(),
                     scopedname = dotName(fscopedName))
 
+        if not self.at_module_scope:
+            self.st.out(struct_class_name, cname = dotName(fscopedName))
+
         mnamel = []
         mdescl = []
         for mem in node.members():
@@ -1378,10 +1513,13 @@ class PythonVisitor:
 
         if len(mnamel) > 0:
             mnames = ", " + string.join(mnamel, ", ")
-            self.st.out(exception_class_init, mnames = mnames)
+        else:
+            mnames = ""
 
-            for mname in mnamel:
-                self.st.out(exception_init_member, mname = mname)
+        self.st.out(exception_class_init, mnames = mnames)
+
+        for mname in mnamel:
+            self.st.out(exception_init_member, mname = mname)
 
         if len(mdescl) > 0:
             mdescs = ", " + string.join(mdescl, ", ")
@@ -1425,6 +1563,9 @@ class PythonVisitor:
                     repoId     = node.repoId(),
                     scopedname = dotName(fscopedName))
 
+        if not self.at_module_scope:
+            self.st.out(union_class_name, cname = dotName(fscopedName))
+
         if node.constrType():
             self.st.inc_indent()
             ams = self.at_module_scope
@@ -1626,28 +1767,273 @@ class PythonVisitor:
                          node.identifier() + "\n")
 
     def visitValueForward(self, node):
-        sys.stderr.write(node.file() + ":" + str(node.line()) + \
-                         ": omniORBpy does not support valuetype\n")
-        sys.exit(1)
+        if self.handleImported(node): return
+
+        vname = mangle(node.identifier())
+
+        self.st.out(value_forward_at_module_scope,
+                    vname=vname, repoId=node.repoId(), modname=self.modname)
+
 
     def visitValueBox(self, node):
-        sys.stderr.write(node.file() + ":" + str(node.line()) + \
-                         ": omniORBpy does not support valuetype\n")
-        sys.exit(1)
+        if self.handleImported(node): return
+
+        boxname = mangle(node.identifier())
+        boxdesc = typeToDescriptor(node.boxedType())
+
+        self.st.out(valuebox, boxname=boxname, repoId=node.repoId(),
+                    boxdesc=boxdesc, modname=self.modname)
+
 
     def visitValueAbs(self, node):
-        sys.stderr.write(node.file() + ":" + str(node.line()) + \
-                         ": omniORBpy does not support valuetype\n")
-        sys.exit(1)
+        if self.handleImported(node): return
+
+        vname = mangle(node.identifier())
+
+        fscopedName = fixupScopedName(node.scopedName(), "")
+        scopedname  = dotName(fscopedName)
+
+        if node.inherits():
+            inheritl = []
+            for i in node.inherits():
+                i = i.fullDecl()
+                inheritl.append(dotName(fixupScopedName(i.scopedName())))
+            
+            inherits = string.join(inheritl, ", ")
+        else:
+            inherits = "_0_CORBA.ValueBase"
+
+        self.st.out(valueabs_class,
+                    vname=vname, scopedname=scopedname, repoId=node.repoId(),
+                    inherits=inherits, modname=self.modname)
+
+        # Declarations within the value
+        if len(node.declarations()) > 0:
+            self.st.inc_indent()
+            self.at_module_scope = 0
+            self.currentScope.append(node.identifier())
+
+            for d in node.declarations():
+                d.accept(self)
+
+            self.currentScope.pop()
+            self.at_module_scope = 1
+            self.st.dec_indent()
+            self.st.out("")
+
+        basedesc = "_0_CORBA.tcInternal.tv_null"
+
+        self.st.out(value_descriptor_at_module_scope,
+                    vname=vname, modifier="_0_CORBA.VM_ABSTRACT",
+                    tbaseids="None", basedesc=basedesc, mdescs="",
+                    modname=self.modname)
+
 
     def visitValue(self, node):
-        sys.stderr.write(node.file() + ":" + str(node.line()) + \
-                         ": omniORBpy does not support valuetype\n")
-        sys.exit(1)
+        if self.handleImported(node): return
+
+        vname = mangle(node.identifier())
+
+        fscopedName = fixupScopedName(node.scopedName(), "")
+        scopedname  = dotName(fscopedName)
+
+        if node.inherits():
+            inheritl = []
+            for i in node.inherits():
+                i = i.fullDecl()
+                inheritl.append(dotName(fixupScopedName(i.scopedName())))
+            
+        else:
+            inheritl = ["_0_CORBA.ValueBase"]
+
+        skeleton_opl = []
+        for i in node.supports():
+            i = i.fullDecl()
+            sn = fixupScopedName(i.scopedName())
+            sn[0] = sn[0] + "__POA"
+            dn = dotName(sn)
+            inheritl.append(dn)
+            skeleton_opl.append(dn)
+
+        inherits = string.join(inheritl, ", ")
+
+        # Go up the chain of inherited interfaces, picking out the
+        # state members
+        members = []
+        ilist   = []
+        cnode   = node
+        
+        while 1:
+            cin = cnode.inherits()
+            if not cin:
+                break
+            i = cin[0].fullDecl()
+            if not isinstance(i, idlast.Value):
+                break
+            ilist.append(i)
+            cnode = i
+
+        ilist.reverse()
+        ilist.append(node)
+        
+        for i in ilist:
+            members.extend(i.statemembers())
+
+        set_argl = []
+
+        for i in range(len(members)):
+            member = members[i]
+            for d in member.declarators():
+                set_argl.append("self.%s = args[%d]" %
+                                (mangle(d.identifier()),i))
+
+        if set_argl:
+            set_args = string.join(set_argl, "\n")
+        else:
+            set_args = "pass"
+
+        if len(set_argl) == 1:
+            s = ""
+        else:
+            s = "s"
+
+        self.st.out(value_class,
+                    vname=vname, scopedname=scopedname, repoId=node.repoId(),
+                    inherits=inherits, set_args=set_args, arglen=len(set_argl),
+                    s=s, modname=self.modname)
+
+        # Declarations within the value
+        if len(node.declarations()) > 0:
+            self.st.inc_indent()
+            self.at_module_scope = 0
+            self.currentScope.append(node.identifier())
+
+            for d in node.declarations():
+                d.accept(self)
+
+            self.currentScope.pop()
+            self.at_module_scope = 1
+            self.st.dec_indent()
+            self.st.out("")
+
+        # Skeleton operation declarations if necessary
+        if node.supports():
+            self.st.out(skeleton_methodmap, methodmap="{}")
+            for i in skeleton_opl:
+                self.st.out(skeleton_inheritmap, inheritclass=i)
+
+            self.st.out(skeleton_set_skel, ifid=vname)
+
+        # Register factory if no callables or factories
+        register_factory = 1
+
+        if not all_factories:
+            cnode = node
+            while 1:
+                if cnode.callables() or cnode.factories() or cnode.supports():
+                    register_factory = 0
+                    break
+                cin = cnode.inherits()
+                if not cin:
+                    break
+                for n in cin:
+                    n = n.fullDecl()
+                    if not isinstance(n, idlast.Value):
+                        register_factory = 0
+                        break
+                cnode = cin[0].fullDecl()
+
+        if register_factory:
+            self.st.out(value_register_factory, vname=vname)
+        
+        # If value supports some interfaces, output an objref class for it
+        if node.supports():
+            inheritl = []
+            methodl  = []
+            for i in node.supports():
+                i = i.fullDecl()
+                sn = fixupScopedName(i.scopedName())
+                inheritl.append(dotName(sn[:-1] + ["_objref_" + sn[-1]]))
+                methodl.append(dotName(sn[:-1] + ["_objref_" + sn[-1]]) +
+                               ".__methods__")
+                
+            inherits = string.join(inheritl, ", ")
+
+            self.st.out(objref_class, ifid=vname, inherits=inherits)
+
+            for inclass in inheritl:
+                self.st.out(objref_inherit_init, inclass=inclass)
+
+            methods = string.join(methodl, " + ")
+            self.st.out(objref_methods, methods = methods)
+
+            # registerObjRef()
+            self.st.out(value_objref_register,
+                        ifid=vname, modname=self.modname)
+
+        # Modifier
+        if node.custom():
+            modifier = "_0_CORBA.VM_CUSTOM"
+        elif node.truncatable():
+            modifier = "_0_CORBA.VM_TRUNCATABLE"
+        else:
+            modifier = "_0_CORBA.VM_NONE"
+
+        # Truncatable bases
+        tbasel = []
+        cnode  = node
+        while 1:
+            cin = cnode.inherits()
+            if not cin:
+                break
+            i = cin[0]
+            i = i.fullDecl()
+            if not isinstance(i, idlast.Value):
+                break
+            if cnode.truncatable():
+                sn = fixupScopedName(i.scopedName())
+                tbasel.append(dotName(sn) + "._NP_RepositoryId")
+            else:
+                break
+            cnode = i
+
+        if tbasel:
+            tbaseids = "(%s._NP_RepositoryId, %s)" % (vname,
+                                                     string.join(tbasel, ", "))
+        else:
+            tbaseids = "None"
+
+        basedesc = None
+        if node.inherits():
+            i = node.inherits()[0].fullDecl()
+            if isinstance(i, idlast.Value):
+                sn = i.scopedName()[:]
+                sn[-1] = "_d_" + sn[-1]
+                basedesc = dotName(fixupScopedName(sn))
+
+        if basedesc is None:
+            basedesc = "_0_CORBA.tcInternal.tv_null"
+
+        mlist = []
+        for m in node.statemembers():
+            for d in m.declarators():
+                mlist.append('"%s"' % mangle(d.identifier()))
+                mlist.append(typeAndDeclaratorToDescriptor(m.memberType(),
+                                                           d, []))
+                if m.memberAccess() == 1:
+                    mlist.append("_0_CORBA.PRIVATE_MEMBER")
+                else:
+                    mlist.append("_0_CORBA.PUBLIC_MEMBER")
+                    
+        mdescs = string.join(mlist, ", ")
+        self.st.out(value_descriptor_at_module_scope,
+                    vname=vname, modifier=modifier, tbaseids=tbaseids,
+                    basedesc=basedesc, mdescs=mdescs, modname=self.modname)
+
 
 
 def docConst(node):
-    if isinstance(node, idlast.Const)     and \
+    if isinstance(node, idlast.Const)        and \
        node.constKind() == idltype.tk_string and \
        node.identifier()[-7:] == "__doc__":
         return node.identifier()[:-7]
@@ -1834,7 +2220,7 @@ class ExampleVisitor (idlvisitor.AstVisitor, idlvisitor.TypeVisitor):
             inheritance_note = ""
 
         for inh in node.inherits():
-            iname = idlutil.ccolonName(inh.scopedName())
+            iname = idlutil.ccolonName(inh.fullDecl().scopedName())
             inheritance_note = inheritance_note + "    #   %s\n" % iname
         
         self.st.out(example_classdef, ifname=ifname,
@@ -1959,24 +2345,30 @@ class ExampleVisitor (idlvisitor.AstVisitor, idlvisitor.TypeVisitor):
 def operationToDescriptors(op):
     """Return the descriptors for an operation.
 
-    Returns a tuple containing (in descriptor, out descriptor,
-    exception map, context list)
+    Returns a tuple containing strings of (in descriptor, out
+    descriptor, exception map, context list, contains values)
     """
 
     indl  = []
     outdl = []
+    cv    = 0
 
     if op.returnType() is not None and \
        op.returnType().kind() != idltype.tk_void:
+
         outdl.append(typeToDescriptor(op.returnType()))
+        cv = idltype.containsValueType(op.returnType())
 
     # Make the lists of in and out parameters
     for p in op.parameters():
+
         if p.is_in():
             indl.append(typeToDescriptor(p.paramType()))
         if p.is_out():
             outdl.append(typeToDescriptor(p.paramType()))
 
+        cv = cv or idltype.containsValueType(p.paramType())
+
     # Fudge single-item lists so that single item tuples work
     if len(indl)  == 1: indl.append("")
     if len(outdl) == 1: outdl.append("")
@@ -2006,7 +2398,7 @@ def operationToDescriptors(op):
     else:
         ctxts = None
 
-    return inds, outds, excs, ctxts
+    return inds, outds, excs, ctxts, cv
 
 
 
@@ -2030,9 +2422,6 @@ ttdMap = {
 
 unsupportedMap = {
     idltype.tk_longdouble: "long double",
-    idltype.tk_value:      "valuetype",
-    idltype.tk_value_box:  "value box",
-    idltype.tk_abstract_interface: "abstract interface"
 }
 
 def typeToDescriptor(tspec, from_scope=[], is_typedef=0):
@@ -2103,13 +2492,22 @@ def dotName(scopedName, our_scope=[]):
 
 def mangle(name):
     if keyword.iskeyword(name): return "_" + name
+
+    # None is a pseudo-keyword that cannot be assigned to.
+    if name == "None": return "_None"
+
     return name
 
 def fixupScopedName(scopedName, prefix="_0_"):
     """Add a prefix and _GlobalIDL to the front of a ScopedName if necessary"""
 
-    if isinstance(idlast.findDecl([scopedName[0]]), idlast.Module):
-        scopedName = [prefix + scopedName[0]] + scopedName[1:]
+    try:
+        decl = idlast.findDecl([scopedName[0]])
+    except idlast.DeclNotFound:
+        decl = None
+
+    if isinstance(decl, idlast.Module):
+        scopedName = [prefix + mangle(scopedName[0])] + scopedName[1:]
     else:
         scopedName = [prefix + global_module] + scopedName
     return scopedName
@@ -2219,7 +2617,7 @@ def real_updateModules(modules, pymodule):
             
             st = output.Stream(f, 4)
 
-            st.out(pymodule_template, module=module)
+            st.out(pymodule_template, module=module, package=module_package)
 
             f.close()
             del f, st
