@@ -28,20 +28,73 @@
 #    omniORB module -- omniORB specific things
 
 
-# $Id: __init__.py,v 1.26.2.19 2003/11/19 14:23:42 dgrisby Exp $
+# $Id: __init__.py,v 1.30.2.21 2008/02/01 16:29:16 dgrisby Exp $
 # $Log: __init__.py,v $
-# Revision 1.26.2.19  2003/11/19 14:23:42  dgrisby
-# Make COS directory a Python package; add it to sys.path if it's not
-# already there.
+# Revision 1.30.2.21  2008/02/01 16:29:16  dgrisby
+# Error with implementation of operations with names clashing with
+# Python keywords.
 #
-# Revision 1.26.2.18  2003/11/10 16:50:29  dgrisby
-# Last checking broke hashability.
+# Revision 1.30.2.20  2007/10/07 15:30:58  dgrisby
+# Problems with modules inside packages. Thanks Fabian Knittel.
 #
-# Revision 1.26.2.17  2003/11/10 16:31:58  dgrisby
-# __cmp__ for enum items.
+# Revision 1.30.2.19  2007/05/11 09:37:23  dgrisby
+# Ensure hash value of unpickled enum items is the same as that of the
+# original item.
 #
-# Revision 1.26.2.16  2003/10/28 13:59:43  dgrisby
-# myIPAddresses function.
+# Revision 1.30.2.18  2006/09/07 15:29:57  dgrisby
+# Use boxes.idl to build standard value boxes.
+#
+# Revision 1.30.2.17  2006/07/11 13:53:09  dgrisby
+# Implement missing TypeCode creation functions.
+#
+# Revision 1.30.2.16  2006/02/22 13:05:15  dgrisby
+# __repr__ and _narrow methods for valuetypes.
+#
+# Revision 1.30.2.15  2006/01/19 17:28:44  dgrisby
+# Merge from omnipy2_develop.
+#
+# Revision 1.30.2.14  2006/01/17 17:38:20  dgrisby
+# Expose omniORB.setClientConnectTimeout function.
+#
+# Revision 1.30.2.13  2005/11/09 12:33:31  dgrisby
+# Support POA LocalObjects.
+#
+# Revision 1.30.2.12  2005/09/01 15:14:41  dgrisby
+# Merge from omnipy3_develop.
+#
+# Revision 1.30.2.11  2005/07/29 11:21:35  dgrisby
+# Fix long-standing problem with module re-opening by #included files.
+#
+# Revision 1.30.2.10  2005/06/24 17:36:00  dgrisby
+# Support for receiving valuetypes inside Anys; relax requirement for
+# old style classes in a lot of places.
+#
+# Revision 1.30.2.9  2005/04/25 18:28:29  dgrisby
+# Minor code for TRANSIENT_FailedOnForwarded.
+#
+# Revision 1.30.2.8  2005/04/14 13:50:45  dgrisby
+# New traceTime, traceInvocationReturns functions; removal of omniORB::logf.
+#
+# Revision 1.30.2.7  2005/03/02 13:39:16  dgrisby
+# Another merge from omnipy2_develop.
+#
+# Revision 1.30.2.6  2005/01/25 11:45:48  dgrisby
+# Merge from omnipy2_develop; set RPM version.
+#
+# Revision 1.30.2.5  2005/01/07 00:22:35  dgrisby
+# Big merge from omnipy2_develop.
+#
+# Revision 1.30.2.4  2003/09/04 14:08:41  dgrisby
+# Correct register_value_factory semantics.
+#
+# Revision 1.30.2.3  2003/07/10 22:13:25  dgrisby
+# Abstract interface support.
+#
+# Revision 1.30.2.2  2003/05/20 17:10:25  dgrisby
+# Preliminary valuetype support.
+#
+# Revision 1.30.2.1  2003/03/23 21:51:43  dgrisby
+# New omnipy3_develop branch.
 #
 # Revision 1.26.2.15  2003/03/12 11:17:49  dgrisby
 # Any / TypeCode fixes.
@@ -196,6 +249,24 @@ import _omnipy
 _coreVersion = _omnipy.coreVersion()
 
 
+# Make sure _omnipy submodules are in sys.modules, and have not been
+# damaged. This can happen if someone has messed with sys.modules, or
+# the interpreter has been stopped and restarted.
+reinit = 0
+for k, v in _omnipy.__dict__.items():
+    if k[-5:] == "_func" and isinstance(v, types.ModuleType):
+        sub = "_omnipy." + k
+        if not sys.modules.has_key(sub):
+            reinit = 1
+            sys.modules[sub] = v
+        del sub
+del k, v
+
+if reinit:
+    _omnipy.ensureInit()
+del reinit
+
+
 # Add path to COS stubs if need be
 _cospath = os.path.join(os.path.dirname(__file__), "COS")
 if _cospath not in sys.path:
@@ -285,8 +356,21 @@ sys.modules."""
             raise ImportError("Error spawning omniidl")
     try:
         m.__file__ = idlname
-        return m._exported_modules
-    except AttributeError:
+        mods = m._exported_modules
+
+        for mod in mods:
+            for m in (mod, skeletonModuleName(mod)):
+                if _partialModules.has_key(m):
+                    if sys.modules.has_key(m):
+                        sys.modules[m].__dict__.update(
+                            _partialModules[m].__dict__)
+                    else:
+                        sys.modules[m] = _partialModules[m]
+                    del _partialModules[m]
+
+        return mods
+
+    except (AttributeError, KeyError):
         del sys.modules[modname]
         raise ImportError("Invalid output from omniidl")
 
@@ -317,33 +401,45 @@ sys.modules."""
     return ret
 
 
-def cdrMarshal(tc, data):
-    """cdrMarshal(TypeCode, data) -> binary string
+def cdrMarshal(tc, data, endian=-1):
+    """cdrMarshal(TypeCode, data [,endian]) -> binary string
 
 Marshal data with the given type into a CDR encapsulation. The data
 can later be converted back into Python objects with cdrUnmarshal().
 The encapsulation is language, platform, and ORB independent.
 
+If the endian boolean is provided, it represents the endianness to
+marshal with: True for little endian; false for big endian. The
+resulting string in this case is the raw marshalled form, not a CDR
+encapsulation. To unmarshal it, the endianness must be known.
+
 Throws CORBA.BAD_PARAM if the data does not match the TypeCode."""
 
     if not isinstance(tc, CORBA.TypeCode):
         raise TypeError("Argument 1 must be a TypeCode")
 
-    return _omnipy.cdrMarshal(tc._d, data)
+    return _omnipy.cdrMarshal(tc._d, data, endian)
 
-def cdrUnmarshal(tc, encap):
-    """cdrUnmarshal(TypeCode, string) -> data
+def cdrUnmarshal(tc, encap, endian=-1):
+    """cdrUnmarshal(TypeCode, string [,endian]) -> data
 
 Unmarshal a CDR stream created with cdrMarshal() or equivalent. The
 encapsulation must adhere to the given TypeCode.
 
+If the endian boolean is provided, it represents the endianness to
+unmarshal with: True for little endian; false for big endian. In this
+case, the string should be the raw marshalled form, not a CDR
+encapsulation. If the endianness does not match that used for
+marshalling, invalid data may be returned, or exceptions raised.
+
 Throws CORBA.MARSHAL if the binary string does not match the
 TypeCode."""
 
     if not isinstance(tc, CORBA.TypeCode):
         raise TypeError("Argument 1 must be a TypeCode")
 
-    return _omnipy.cdrUnmarshal(tc._d, encap)
+    return _omnipy.cdrUnmarshal(tc._d, encap, endian)
+
 
 WTHREAD_CREATED = 0
 WTHREAD_DELETED = 1
@@ -376,14 +472,19 @@ Make stubs for the Interface Repository appear in the CORBA module"""
 #   installSystemExceptionHandler()
 #   traceLevel
 #   traceInvocations
+#   traceInvocationReturns
 #   traceThreadId
+#   traceTime
 #   log
 #   nativeCharCodeSet
 #   fixed
 #   minorCodeToString
 #   setClientCallTimeout
 #   setClientThreadCallTimeout
+#   setClientConnectTimeout
 #   myIPAddresses
+#   setPersistentServerIdentifier
+#   locationForward
 
 from _omnipy.omni_func import *
 
@@ -393,19 +494,25 @@ from _omnipy.omni_func import *
 # Private things
 
 # ORB:
-orb     = None
-rootPOA = None
-lock    = threading.Lock()
+orb      = None
+rootPOA  = None
+poaCache = {}
+lock     = threading.Lock()
 
 # Maps for object reference classes and IDL-defined types:
-objrefMapping   = {}
-typeMapping     = {}
-typeCodeMapping = {}
+objrefMapping       = {}
+skeletonMapping     = {}
+typeMapping         = {}
+typeCodeMapping     = {}
+valueFactoryMapping = {}
 
 
 def registerObjref(repoId, objref):
     objrefMapping[repoId] = objref
 
+def registerSkeleton(repoId, skel):
+    skeletonMapping[repoId] = skel
+
 def registerType(repoId, desc, tc):
     typeMapping[repoId]     = desc
     typeCodeMapping[repoId] = tc
@@ -416,13 +523,38 @@ def findType(repoId):
 def findTypeCode(repoId):
     return typeCodeMapping.get(repoId)
 
+def registerValueFactory(repoId, factory):
+    old = valueFactoryMapping.get(repoId)
+    valueFactoryMapping[repoId] = factory
+    return old
+
+def unregisterValueFactory(repoId):
+    del valueFactoryMapping[repoId]
+
+def findValueFactory(repoId):
+    return valueFactoryMapping.get(repoId)
+
+
+# Map of partially-opened modules
+_partialModules = {}
+
 
 # Function to return a Python module for the required IDL module name
 def openModule(mname, fname=None):
     if mname == "CORBA":
         mod = sys.modules["omniORB.CORBA"]
+
     elif sys.modules.has_key(mname):
         mod = sys.modules[mname]
+
+        if _partialModules.has_key(mname):
+            pmod = _partialModules[mname]
+            mod.__dict__.update(pmod.__dict__)
+            del _partialModules[mname]
+            
+    elif _partialModules.has_key(mname):
+        mod = _partialModules[mname]
+
     else:
         mod = newModule(mname)
 
@@ -447,15 +579,34 @@ def newModule(mname):
 
         if sys.modules.has_key(current):
             mod = sys.modules[current]
+
+        elif _partialModules.has_key(current):
+            mod = _partialModules[current]
+
         else:
             newmod = imp.new_module(current)
-            if mod: setattr(mod, name, newmod)
-            sys.modules[current] = mod = newmod
+            _partialModules[current] = mod = newmod
 
         current = current + "."
 
     return mod
 
+# Function to update a module with the partial module store in the
+# partial module map
+def updateModule(mname):
+    if _partialModules.has_key(mname):
+        pmod = _partialModules[mname]
+        mod  = sys.modules[mname]
+        mod.__dict__.update(pmod.__dict__)
+        del _partialModules[mname]
+
+
+def skeletonModuleName(mname):
+    l = string.split(mname, ".")
+    l[0] = l[0] + "__POA"
+    return string.join(l, ".")
+
+
 
 # Function to create a new empty class as a scope place-holder
 def newEmptyClass():
@@ -478,10 +629,19 @@ class EnumItem:
         return self._n
 
     def __cmp__(self, other):
-        return cmp(self._v, other._v)
+        try:
+            if isinstance(other, EnumItem):
+                if other._parent_id == self._parent_id:
+                    return cmp(self._v, other._v)
+                else:
+                    return cmp(self._parent_id, other._parent_id)
+            else:
+                return cmp(id(self), id(other))
+        except:
+            return cmp(id(self), id(other))
 
     def __hash__(self):
-        return id(self)
+        return hash(self._parent_id + "/" + self._n)
 
 class AnonymousEnumItem (EnumItem):
     def __init__(self, value):
@@ -503,7 +663,45 @@ class Enum:
         return self._items[n]
 
 
+class StructBase:
+    _NP_RepositoryId = None
+    _NP_ClassName = None
+    
+    def __repr__(self):
+        cname = self._NP_ClassName
+        if cname is None:
+            cname = "%s.%s" % (self.__module__, self.__class__.__name__)
+
+        desc = findType(self._NP_RepositoryId)
+        if desc is None:
+            # Type is not properly registered
+            return "<%s instance at 0x%x>" % (cname, id(self))
+        vals = []
+        for i in range(4, len(desc), 2):
+            attr = desc[i]
+            try:
+                val = getattr(self, attr)
+                vals.append("%s=%s" % (attr,repr(val)))
+            except AttributeError:
+                vals.append("%s=<not set>" % attr)
+
+        return "%s(%s)" % (cname, string.join(vals, ", "))
+
+    def _tuple(self):
+        desc = findType(self._NP_RepositoryId)
+        if desc is None:
+            # Type is not properly registered
+            raise CORBA.BAD_PARAM(BAD_PARAM_IncompletePythonType,
+                                  CORBA.COMPLETED_NO)
+        vals = []
+        for i in range(4, len(desc), 2):
+            attr = desc[i]
+            vals.append(getattr(self, attr))
+        return tuple(vals)
+
+
 class Union:
+    _NP_ClassName = None
     _def_m = None
 
     def __init__(self, *args, **kw):
@@ -513,21 +711,10 @@ class Union:
         else:
             ks = kw.keys()
             if len(args) != 0 or len(ks) != 1:
-                raise AttributeError("require 2 arguments or one keyword argument.")
+                raise TypeError("require 2 arguments or one keyword argument.")
             k = ks[0]
             self.__setattr__(k, kw[k])
 
-    def __str__(self):
-        try:
-            mem = self._d_to_m[self._d]
-            return "_d = " + str(self._d) + ", " + mem + " = " + str(self._v)
-        except KeyError:
-            if self._def_m is not None:
-                return "_d = " + str(self._d) + ", " + self._def_m + \
-                       " = " + str(self._v)
-            else:
-                return "_d = " + str(self._d)
-
     def __getattr__(self, mem):
         try:
             cmem = self._d_to_m[self._d]
@@ -564,21 +751,29 @@ class Union:
                 else:
                     raise AttributeError(mem)
 
+    def __repr__(self):
+        cname = self._NP_ClassName
+        if cname is None:
+            cname = "%s.%s" % (self.__module__, self.__class__.__name__)
+
+        try:
+            return "%s(%s = %s)" % (cname, self._d_to_m[self._d],
+                                    repr(self._v))
+        except KeyError:
+            return "%s(%s, %s)" % (cname, repr(self._d), repr(self._v))
+
+
 # Import sub-modules
 import CORBA, tcInternal
 
 def createUnknownStruct(repoId, members):
 
-    class UnknownStruct:
+    class UnknownStruct (StructBase):
         def __init__(self, *args):
-            if len(args) < len(self._members):
-                raise TypeError("not enough arguments; expected " + \
-                                str(len(self._members)) + ", got " + \
-                                str(len(args)))
-            elif len(args) > len(self._members):
-                raise TypeError("too many arguments; expected " + \
-                                str(len(self._members)) + ", got " + \
-                                str(len(args)))
+            if len(args) != len(self._members):
+                raise TypeError("__init__() takes exactly %d arguments "
+                                "(%d given)" %
+                                (len(self._members) + 1, len(args) + 1))
 
             self._values = args
 
@@ -586,6 +781,21 @@ def createUnknownStruct(repoId, members):
                 if self._members[i] != "":
                     setattr(self, self._members[i], args[i])
 
+        def __repr__(self):
+            vals = []
+            for i in range(len(self._values)):
+                attr = self._members[i]
+                val  = self._values[i]
+                if attr:
+                    vals.append("%s=%s" % (attr, repr(val)))
+                else:
+                    vals.append(repr(val))
+
+            return "UnknownStruct<%s>(%s)" % (self._NP_RepositoryId,
+                                              string.join(vals, ", "))
+        def _tuple(self):
+            return tuple(self._values)
+
     UnknownStruct._NP_RepositoryId = repoId
     UnknownStruct._members         = members
     return UnknownStruct
@@ -596,6 +806,7 @@ def createUnknownUnion(repoId, def_used, members):
         pass
 
     UnknownUnion._NP_RepositoryId = repoId
+    UnknownUnion._NP_ClassName    = "UnknownUnion<%s>" % repoId
     UnknownUnion._d_to_m          = {}
     UnknownUnion._m_to_d          = {}
 
@@ -613,14 +824,10 @@ def createUnknownUserException(repoId, members):
 
     class UnknownUserException (CORBA.UserException):
         def __init__(self, *args):
-            if len(args) < len(self._members):
-                raise TypeError("not enough arguments; expected " + \
-                                str(len(self._members)) + ", got " + \
-                                str(len(args)))
-            elif len(args) > len(self._members):
-                raise TypeError("too many arguments; expected " + \
-                                str(len(self._members)) + ", got " + \
-                                str(len(args)))
+            if len(args) != len(self._members):
+                raise TypeError("__init__() takes exactly %d arguments "
+                                "(%d given)" %
+                                (len(self._members) + 1, len(args) + 1))
 
             self._values = args
 
@@ -628,11 +835,46 @@ def createUnknownUserException(repoId, members):
                 if self._members[i] != "":
                     setattr(self, self._members[i], args[i])
 
+        def __repr__(self):
+            vals = []
+            for i in range(len(self._values)):
+                attr = self._members[i]
+                val  = self._values[i]
+                if attr:
+                    vals.append("%s=%s" % (attr, repr(val)))
+                else:
+                    vals.append(repr(val))
+
+            return "UnknownUserException<%s>(%s)" % (self._NP_RepositoryId,
+                                                     string.join(vals, ", "))
+
     UnknownUserException._NP_RepositoryId = repoId
     UnknownUserException._members         = members
     return UnknownUserException
 
 
+class UnknownValueBase (CORBA.ValueBase):
+    pass
+
+
+def createUnknownValue(repoId, base_desc):
+
+    if base_desc == tcInternal.tv_null:
+        class UnknownValue (UnknownValueBase):
+            pass
+    else:
+        base_cls = base_desc[1]
+        if isinstance(base_cls, UnknownValueBase):
+            class UnknownValue (base_cls):
+                pass
+        else:
+            class UnknownValue (UnknownValueBase, base_cls):
+                pass
+
+    UnknownValue._NP_RepositoryId = repoId
+    return UnknownValue
+
+
 # Function to coerce an Any value with a partially-specified
 # descriptor to a value with an equivalent, fully-specified
 # descriptor.
@@ -747,15 +989,16 @@ class WorkerThread(threading.Thread):
     hooks = []
 
     def __init__(self):
-        _thr_init(self, name="omniORB")
+        id = _thr_id()
+        _thr_init(self, name="omniORB-%d" % id)
         self._Thread__started = 1
-        self.id = _thr_id()
+        self.id = id
         _thr_acq()
-        if _thr_act.has_key(self.id):
+        if _thr_act.has_key(id):
             self.add = 0
         else:
             self.add = 1
-            _thr_act[self.id] = self
+            _thr_act[id] = self
         _thr_rel()
         if self.add:
             for hook in self.hooks:
@@ -766,8 +1009,10 @@ class WorkerThread(threading.Thread):
             for hook in self.hooks:
                 hook(WTHREAD_DELETED, self)
             _thr_acq()
-            del _thr_act[self.id]
-            _thr_rel()
+            try:
+                del _thr_act[self.id]
+            finally:
+                _thr_rel()
 
     def _set_daemon(self): return 1
     def join(self):        assert 0, "cannot join an omniORB WorkerThread"
@@ -809,6 +1054,9 @@ try:
     import keyword
     for word in keyword.kwlist:
         keywordMapping[word] = "_" + word
+
+    keywordMapping["None"] = "_None"
+
     del keyword
 except ImportError:
     pass
@@ -821,16 +1069,7 @@ def omniORBminorCode(c):
 def OMGminorCode(c):
     return 0x4f4d0000 | c
 
-BAD_INV_ORDER_ORBHasShutdown       = OMGminorCode(4)
-BAD_CONTEXT_NoMatchingProperty     = omniORBminorCode(75)
-BAD_CONTEXT_StartingScopeNotFound  = omniORBminorCode(76)
-BAD_PARAM_WrongPythonType          = omniORBminorCode(88)
-BAD_PARAM_PythonValueOutOfRange    = omniORBminorCode(95)
-BAD_PARAM_WrongUnionMemberSelected = omniORBminorCode(107)
-BAD_PARAM_InvalidFixedPointLimits  = omniORBminorCode(109)
-BAD_TYPECODE_InvalidIndirection    = omniORBminorCode(108)
-INTF_REPOS_NotAvailable            = omniORBminorCode(39)
-NO_IMPLEMENT_Unsupported           = omniORBminorCode(36)
+from omniORB.minorCodes import *
 
 
 # More public things, which depend on the CORBA module
@@ -855,14 +1094,33 @@ is set to 1, a permanent location forward is requested."""
         self._perm    = perm
 
     def __str__(self):
-        return "Location forward exception"
+        return "omniORB.LOCATION_FORWARD exception"
 
+# "Static" objects required by the _omnipy module. They are here so
+# memory management works correctly if the omniORB modules are
+# unloaded.
+
+_emptyTuple      = ()
+_ORB_TWIN        = "__omni_orb"
+_OBJREF_TWIN     = "__omni_obj"
+_SERVANT_TWIN    = "__omni_svt"
+_POA_TWIN        = "__omni_poa"
+_POAMANAGER_TWIN = "__omni_mgr"
+_POACURRENT_TWIN = "__omni_pct"
+_NP_RepositoryId = "_NP_RepositoryId"
 
 
 # Register this module and the threading module with omnipy:
 import omniORB, omniORB.PortableServer
 _omnipy.registerPyObjects(omniORB)
 
+# Import CORBA module stubs
+import corbaidl_idl
+import boxes_idl
+
+sys.modules["corbaidl_idl"] = corbaidl_idl
+sys.modules["boxes_idl"]    = boxes_idl
+
 # Import the Interface Repository stubs if necessary
 if os.environ.has_key("OMNIORBPY_IMPORT_IR_STUBS"):
     importIRStubs()
